"""CLI package."""

import logging
import os

from cyclopts import App

from ..config import ConfigModel
from ..config.converter import config_to_profile_file, profile_to_config_file
from ..core import build_engine
from ..utils import DefaultPath, Paths
from .observer import ConsoleObserver

logger = logging.getLogger(__name__)
app = App(
    help="fspachinko - Random file transfer utility.",
)


@app.default
def run(config: str = "") -> None:
    """Run the fspachinko CLI.

    Args:
        config (str): Path to configuration file.

    """
    if not config:
        config = Paths.config(DefaultPath.CONFIG)

    try:
        with open(config, encoding="utf-8") as f:
            data = f.read()
    except FileNotFoundError:
        logger.exception("Configuration file not found: %s", config)
        return

    observer = ConsoleObserver()
    config_model = ConfigModel.model_validate_json(data)
    engine = build_engine(config_model)
    engine.set_observer(observer)
    engine.start()


@app.command
def profile_to_config(profile: str, output: str) -> None:
    """Convert a GUI profile JSON to a fspachinko config JSON."""
    if not os.path.exists(profile):
        logger.error("Profile file not found: %s", profile)
        return

    profile_to_config_file(profile, output)


@app.command
def config_to_profile(config: str, output: str) -> None:
    """Convert a fspachinko config JSON to a GUI profile JSON."""
    if not os.path.exists(config):
        logger.error("Config file not found: %s", config)
        return

    config_to_profile_file(config, output)



"""Main module."""

import logging
from typing import TYPE_CHECKING

from PySide6.QtCore import QSettings, Slot
from PySide6.QtWidgets import QFileDialog, QMainWindow, QStatusBar, QToolBar

from ..utils import GUIFileDialogFilter, GUILabel, GUIName, GUISettingsKey, GUITitle, get_stem_and_ext
from .actions import Actions
from .centralwidget import CentralWidget
from .qthelpers import set_qt_name
from .settings import ProfileManager

if TYPE_CHECKING:
    from PySide6.QtGui import QCloseEvent

logger = logging.getLogger(__name__)


class MainWindow(QMainWindow):
    """Main application window."""

    def __init__(self) -> None:
        """Initialize the main window."""
        super().__init__()
        logger.debug("Initializing GUI")
        self.central_widget = CentralWidget()
        self.profiles = ProfileManager()
        self.qsettings = QSettings()
        self._actions = Actions()
        self.setCentralWidget(self.central_widget)
        self.init_connections()
        self.init_menubar()
        self.init_toolbar()
        self.init_statusbar()
        self.init_settings()

    def init_connections(self) -> None:
        """Initialize connections."""
        self._actions.file.save.triggered.connect(self.save_profile)
        self._actions.file.save_as.triggered.connect(self.save_profile_as_dialog)
        self._actions.file.load.triggered.connect(self.open_profile_dialog)
        self._actions.file.exit.triggered.connect(self.close)

        self._actions.run.start.triggered.connect(self.central_widget.on_start)
        self._actions.run.stop.triggered.connect(self.central_widget.on_stop)

    def init_menubar(self) -> None:
        """Initialize the menu bar."""
        menubar = self.menuBar()
        set_qt_name(menubar, GUIName.MENUBAR)

        file_menu = menubar.addMenu(GUILabel.FILEMENU)
        set_qt_name(file_menu, GUIName.FILEMENU)

        file_menu.addAction(self._actions.file.save)
        file_menu.addAction(self._actions.file.save_as)
        file_menu.addAction(self._actions.file.load)
        file_menu.addSeparator()
        file_menu.addAction(self._actions.file.autosave)
        file_menu.addSeparator()
        file_menu.addAction(self._actions.file.exit)

        run_menu = menubar.addMenu(GUILabel.RUNMENU)
        set_qt_name(run_menu, GUIName.RUNMENU)

        run_menu.addAction(self._actions.run.start)
        run_menu.addAction(self._actions.run.stop)

    def init_toolbar(self) -> None:
        """Initialize the toolbar."""
        toolbar = QToolBar(GUIName.TOOLBAR)
        set_qt_name(toolbar, GUIName.TOOLBAR)

        toolbar.addAction(self._actions.file.save)
        toolbar.addAction(self._actions.file.save_as)
        toolbar.addAction(self._actions.file.load)
        toolbar.addAction(self._actions.file.autosave)
        toolbar.addSeparator()
        toolbar.addAction(self._actions.run.start)
        toolbar.addAction(self._actions.run.stop)
        toolbar.addSeparator()
        toolbar.addAction(self._actions.file.exit)

        self.addToolBar(toolbar)

    def init_statusbar(self) -> None:
        """Initialize the status bar."""
        statusbar = QStatusBar(self, sizeGripEnabled=True)
        set_qt_name(statusbar, GUIName.STATUSBAR)

        self.setStatusBar(statusbar)

    def init_settings(self) -> None:
        """Initialize GUI settings manager."""
        self.restoreGeometry(self.qsettings.value(GUISettingsKey.GEOMETRY))
        self.restoreState(self.qsettings.value(GUISettingsKey.STATE))
        self.profiles.set_current(str(self.qsettings.value(GUISettingsKey.PROFILE, "")))
        self.profiles.open_profile(self)
        self.reset_window_title()

    @Slot()
    def save_profile(self) -> None:
        """Save the current GUI profile."""
        self.profiles.save_profile(self)

    @Slot()
    def save_profile_as_dialog(self) -> None:
        """Save a GUI profile via dialog."""
        filename, _ = QFileDialog.getSaveFileName(
            self,
            GUITitle.SAVE_PROFILE,
            self.profiles.get_current_profile_parent(),
            GUIFileDialogFilter.JSON,
        )
        if filename:
            self.profiles.set_current(filename)
            self.save_profile()
            self.reset_window_title()

    @Slot()
    def open_profile_dialog(self) -> None:
        """Load a GUI profile via dialog."""
        filename, _ = QFileDialog.getOpenFileName(
            self,
            GUITitle.OPEN_PROFILE,
            self.profiles.get_current_profile_parent(),
            GUIFileDialogFilter.JSON,
        )
        if filename:
            self.profiles.set_current(filename)
            self.profiles.open_profile(self)
            self.reset_window_title()

    @Slot()
    def reset_window_title(self) -> None:
        """Update the window title based on the current profile."""
        profile_stem, _ = get_stem_and_ext(self.profiles.current_profile)
        self.setWindowTitle(f"{profile_stem} - {GUITitle.WINDOW}")

    def save_settings(self) -> None:
        """Save GUI settings on close."""
        self.qsettings.setValue(GUISettingsKey.GEOMETRY, self.saveGeometry())
        self.qsettings.setValue(GUISettingsKey.STATE, self.saveState())
        self.qsettings.setValue(GUISettingsKey.PROFILE, self.profiles.current_profile)

    def closeEvent(self, event: QCloseEvent) -> None:  # noqa: N802
        """Handle window close event."""
        self.save_settings()
        if self._actions.file.autosave.isChecked():
            self.save_profile()
        super().closeEvent(event)



"""Constants."""

import os
from enum import IntEnum, StrEnum

from .paths import Paths

# Ensure necessary directories exist
os.makedirs(Paths.profiles, exist_ok=True)

# General constants
WALKER_CACHE_LIMIT: int = 1000
PERCENTAGE_100: float = 100.0
INVALID_FILENAME_CHARS: set[str] = set(r'\/:*?"<>|')
TRUE_STRS: set[str] = {"y", "yes", "t", "true", "on", "1"}
FALSE_STRS: set[str] = {"n", "no", "f", "false", "off", "0"}
DURATION_CMD = [
    "ffprobe",
    "-v",
    "error",
    "-show_entries",
    "format=duration",
    "-of",
    "default=noprint_wrappers=1:nokey=1",
]


class DefaultPath(StrEnum):
    """Enumeration for default configuration filenames."""

    CONFIG = "fspachinko.json"
    LOGGING = "logging.json"


class ReStrFmt(StrEnum):
    """Enumeration for regex string formats."""

    KEYWORD = r"(.*){}(.*)"
    EXTENSION = r".{}$"


class FileError(IntEnum):
    """Enumeration for file error codes."""

    WINDOWS_CROSS_DRIVE_ERROR = 17
    UNIX_CROSS_FILESYSTEM_ERROR = 18


class SecondsIn(IntEnum):
    """Enumeration for seconds in units."""

    SECOND = 1
    MINUTE = 60
    HOUR = 3600


class BytesIn(IntEnum):
    """Enumeration for bytes in units."""

    BYTE = 1
    KILOBYTE = 1 << 10
    MEGABYTE = 1 << 20
    GIGABYTE = 1 << 30


class TransferMode(StrEnum):
    """Enumeration for file transfer modes."""

    COPY = "Copy"
    COPY_PRESERVE = "Copy (Preserve)"
    MOVE = "Move"
    SYMLINK = "Symlink"
    HARDLINK = "Hardlink"


class AppSetting(StrEnum):
    """Enumeration for different settings categories."""

    ORGANIZATION = "Wonyoung Jang"
    DOMAIN = "https://github.com/wonyoung-jang/fspachinko"
    APPLICATION = "fspachinko"


class ByteUnit(StrEnum):
    """Enumeration for size units."""

    BYTES = "B"
    KILOBYTES = "KB"
    MEGABYTES = "MB"
    GIGABYTES = "GB"


class TimeUnit(StrEnum):
    """Enumeration for time units."""

    SECONDS = "s"
    MINUTES = "m"
    HOURS = "h"


SIZE_MAP = {
    ByteUnit.BYTES: BytesIn.BYTE,
    ByteUnit.KILOBYTES: BytesIn.KILOBYTE,
    ByteUnit.MEGABYTES: BytesIn.MEGABYTE,
    ByteUnit.GIGABYTES: BytesIn.GIGABYTE,
}

TIME_MAP = {
    TimeUnit.SECONDS: SecondsIn.SECOND,
    TimeUnit.MINUTES: SecondsIn.MINUTE,
    TimeUnit.HOURS: SecondsIn.HOUR,
}


class FilenameTemplate(StrEnum):
    """Enumeration for filename templates."""

    ORIGINAL = "{original}"
    INDEX = "{index}"
    DATE = "{date}"
    TIME = "{time}"
    DATETIME = "{datetime}"
    PARENT = "{parent}"
    PARENTS_TO_ROOT = "{parentstoroot}"


class FilenameTemplateMapKey(StrEnum):
    """Enumeration for filename templates."""

    ORIGINAL = "original"
    INDEX = "index"
    DATE = "date"
    TIME = "time"
    DATETIME = "datetime"
    PARENT = "parent"
    PARENTS_TO_ROOT = "parentstoroot"


class IconFilename(StrEnum):
    """Enumeration for icon filenames."""

    WINDOW = "windowIcon.png"
    SAVE = "save_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    SAVE_AS = "save_as_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    OPEN = "file_open_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    AUTOSAVE = "sync_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    START = "play_arrow_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    STOP = "stop_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    CLOSE = "close_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    BROWSE = "add_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    OPEN_DIR = "folder_open_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"
    REMOVE = "remove_24dp_E3E3E3_FILL0_wght400_GRAD0_opsz24.svg"


class DateTimeFormat(StrEnum):
    """Enumeration for date and time formats."""

    DATE = "%Y-%m-%d"
    TIME = "%H-%M-%S"
    DATETIME = "%Y-%m-%d--%H-%M-%S"


class StateStatus(StrEnum):
    """Enumeration for engine state statuses."""

    USER_STOPPED = "USER STOPPED"
    SUCCESS = "SUCCESS"
    ALL_FILES_SEARCHED = "ALL FILES SEARCHED"
    FOLDER_SIZE_LIMIT_REACHED = "FOLDER SIZE LIMIT REACHED"
    TOTAL_SIZE_LIMIT_REACHED = "TOTAL SIZE LIMIT REACHED"
    NO_FILES_FOUND_ALL_SEARCHED_FOLDER_DELETED = "NO FILES FOUND | ALL FILES SEARCHED | FOLDER DELETED"
    NO_FILES_FOUND_FOLDER_DELETED = "NO FILES FOUND | FOLDER DELETED"


class GUISettingsKey(StrEnum):
    """Enumeration for QSettings keys."""

    GEOMETRY = "geometry"
    STATE = "state"
    PROFILE = "profile"


class GUITitle(StrEnum):
    """Enumeration for GUI window titles."""

    WINDOW = "fspachinko: Transfer random files"
    SAVE_PROFILE = "Save Profile As"
    OPEN_PROFILE = "Open Profile"


class GUIName(StrEnum):
    """Enumeration for GUI object names."""

    CENTRAL_WIDGET = "central_widget"
    MENUBAR = "menubar"
    RUNMENU = "run_menu"
    FILEMENU = "file_menu"
    TOOLBAR = "toolbar"
    STATUSBAR = "statusbar"


class GUILabel(StrEnum):
    """Enumeration for GUI labels."""

    FILEMENU = "&File"
    RUNMENU = "&Run"


class GUIFileDialogFilter(StrEnum):
    """Enumeration for GUI file dialog filters."""

    JSON = "JSON Files (*.json)"



"""Config package."""

from .config import (
    Filecount,
    Filename,
    Folder,
    ListIncludeExclude,
    MinMax,
    SizeLimit,
)
from .schemas import (
    ConfigModel,
    DirectoryModel,
    FilecountModel,
    FilenameModel,
    ListIncludeExcludeModel,
    MinMaxModel,
    OptionsModel,
    SizeLimitModel,
    TransferModeModel,
)

__all__ = [
    "ConfigModel",
    "DirectoryModel",
    "Filecount",
    "FilecountModel",
    "Filename",
    "FilenameModel",
    "Folder",
    "ListIncludeExclude",
    "ListIncludeExcludeModel",
    "MinMax",
    "MinMaxModel",
    "OptionsModel",
    "SizeLimit",
    "SizeLimitModel",
    "TransferModeModel",
]



"""Observer for GUI."""

from dataclasses import dataclass

from PySide6.QtCore import QObject, Signal

from ..utils import Observer


class WorkerSignals(QObject):
    """Qt worker signals."""

    progress_total = Signal(int)
    count_total = Signal()
    progress = Signal(int)
    finished = Signal()
    log = Signal(str)
    time = Signal()
    count = Signal(int)


@dataclass(slots=True)
class GuiObserver(Observer):
    """GUI observer."""

    signals: WorkerSignals

    def on_progress_total(self, maximum: int) -> None:
        """Emit total progress signal."""
        self.signals.progress_total.emit(maximum)

    def on_count_total(self) -> None:
        """Emit total count signal."""
        self.signals.count_total.emit()

    def on_progress(self, maximum: int) -> None:
        """Emit progress signal."""
        self.signals.progress.emit(maximum)

    def on_finished(self) -> None:
        """Emit finished signal."""
        self.signals.finished.emit()

    def on_log(self, msg: str) -> None:
        """Emit log message signal."""
        self.signals.log.emit(msg)

    def on_time(self) -> None:
        """Emit time update signal."""
        self.signals.time.emit()

    def on_count(self, count: int) -> None:
        """Emit count update signal."""
        self.signals.count.emit(count)



"""Utilities package."""

from .constants import (
    DURATION_CMD,
    FALSE_STRS,
    INVALID_FILENAME_CHARS,
    PERCENTAGE_100,
    SIZE_MAP,
    TIME_MAP,
    TRUE_STRS,
    WALKER_CACHE_LIMIT,
    AppSetting,
    BytesIn,
    ByteUnit,
    DefaultPath,
    FileError,
    FilenameTemplate,
    FilenameTemplateMapKey,
    GUIFileDialogFilter,
    GUILabel,
    GUIName,
    GUISettingsKey,
    GUITitle,
    IconFilename,
    ReStrFmt,
    StateStatus,
    TimeUnit,
    TransferMode,
)
from .helpers import (
    SafeDict,
    are_paths_equal,
    calc_unique_path_name,
    convert_byte_to_human_readable_size,
    convert_string_to_list,
    get_duration,
    get_stem_and_ext,
    load_json,
    remove_directory,
    save_json,
    strtobool,
)
from .interfaces import Observer
from .loggers import initialize_logging
from .paths import Paths
from .timestamp import DateTimeStamp

__all__ = [
    "DURATION_CMD",
    "FALSE_STRS",
    "INVALID_FILENAME_CHARS",
    "PERCENTAGE_100",
    "SIZE_MAP",
    "TIME_MAP",
    "TRUE_STRS",
    "WALKER_CACHE_LIMIT",
    "AppSetting",
    "ByteUnit",
    "BytesIn",
    "DateTimeStamp",
    "DefaultPath",
    "FileError",
    "FilenameTemplate",
    "FilenameTemplateMapKey",
    "GUIFileDialogFilter",
    "GUILabel",
    "GUIName",
    "GUISettingsKey",
    "GUITitle",
    "IconFilename",
    "Observer",
    "Paths",
    "ReStrFmt",
    "SafeDict",
    "StateStatus",
    "TimeUnit",
    "TransferMode",
    "are_paths_equal",
    "calc_unique_path_name",
    "convert_byte_to_human_readable_size",
    "convert_string_to_list",
    "get_duration",
    "get_stem_and_ext",
    "initialize_logging",
    "load_json",
    "remove_directory",
    "save_json",
    "strtobool",
]



"""Engine state classes."""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from time import perf_counter
from typing import TYPE_CHECKING

from ..utils import DateTimeStamp, StateStatus, remove_directory

if TYPE_CHECKING:
    from ..config import Folder, SizeLimit
    from .quota import DiversityQuota
    from .reporter import ReportWriter
    from .walker import FSEntry

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class FolderStats:
    """Dataclass for state."""

    count: int = 0
    starttime: float = 0.0
    curr_size: int = 0
    total_size: int = 0

    def reset(self) -> None:
        """Reset state variables for a new folder."""
        self.count = 0
        self.curr_size = 0
        self.starttime = perf_counter()

    def update(self, size: int) -> None:
        """Update state on successful operation."""
        self.count += 1
        self.curr_size += size
        self.total_size += size


@dataclass(slots=True)
class Context(ABC):
    """Abstract base class for engine state context."""

    folder: Folder
    quota: DiversityQuota
    folder_size_limit: SizeLimit
    total_size_limit: SizeLimit
    reporter: ReportWriter
    is_dry_run: bool
    dtstamp: DateTimeStamp
    is_stop_requested: bool = False
    folderstats: FolderStats = field(default_factory=FolderStats)
    _state: EngineState | None = None

    @property
    def state(self) -> EngineState:
        """Get the current engine state."""
        if self._state is None:
            msg = "Engine state is not set."
            raise ValueError(msg)
        return self._state

    @state.setter
    def state(self, new_state: EngineState) -> None:
        """Set a new engine state."""
        self._state = new_state
        self._state.context = self

    @abstractmethod
    def should_stop(self, target: int) -> bool:
        """Check and update state before file validation."""

    @abstractmethod
    def is_none_found(self) -> bool:
        """Check if no files were found in the current folder."""

    @abstractmethod
    def prepare(self, dest: str) -> None:
        """Prepare the context for a new folder processing."""

    @abstractmethod
    def update_on_success(self, entry: FSEntry) -> None:
        """Update context on successful file operation."""

    @abstractmethod
    def should_treat_as_dry_run(self, copy_path_str: str) -> bool:
        """Check if a file has already been transferred."""

    @abstractmethod
    def set_errored(self, copy_path_str: str) -> None:
        """Set the state to invalid file transfer."""

    @abstractmethod
    def set_transferred(self, copy_path_str: str) -> None:
        """Set the state to successful file transfer."""

    @abstractmethod
    def finalize(self, target: int, dest: str) -> str:
        """Finalize the context after processing."""


@dataclass(slots=True)
class EngineContext(Context):
    """Class for engine state context."""

    def should_stop(self, target: int) -> bool:
        """Check and update state before file validation."""
        if self.is_stop_requested:
            self.state = UserStoppedState(
                status=StateStatus.USER_STOPPED,
                message="Stopped by user request",
            )
            return True

        if self.folderstats.count == target:
            self.state = SuccessState(
                status=StateStatus.SUCCESS,
                message=f"Copied {self.folderstats.count}/{target} files",
            )
            return True

        if self.quota.is_all_locked():
            self.state = AllSearched(
                status=StateStatus.ALL_FILES_SEARCHED,
                message="All files locked by diversity quota",
            )
            return True

        if (
            self.folder_size_limit.is_enabled
            and self.folder_size_limit.size_limit > 0
            and self.folder_size_limit.is_valid(self.folderstats.curr_size)
        ):
            self.state = FolderSizeLimitState(
                status=StateStatus.FOLDER_SIZE_LIMIT_REACHED,
                message=f"{(self.folderstats.curr_size)} B / {(self.folder_size_limit.size_limit)} B",
            )
            return True

        if (
            self.total_size_limit.is_enabled
            and self.total_size_limit.size_limit > 0
            and self.total_size_limit.is_valid(self.folderstats.total_size)
        ):
            self.state = TotalSizeLimitState(
                status=StateStatus.TOTAL_SIZE_LIMIT_REACHED,
                message=f"{(self.folderstats.total_size)} B / {(self.total_size_limit.size_limit)} B",
            )
            return True

        return False

    def is_none_found(self) -> bool:
        """Check if no files were found in the current folder."""
        none_found = self.folderstats.count == 0 and self.folder.is_enabled
        if none_found:
            if self.quota.is_all_locked():
                self.state = NoFilesFoundAllSearchedState(
                    status=StateStatus.NO_FILES_FOUND_ALL_SEARCHED_FOLDER_DELETED,
                    message="No files found and all files locked by diversity quota",
                )
                return True

            self.state = NoFilesFoundState(
                status=StateStatus.NO_FILES_FOUND_FOLDER_DELETED,
                message="No files found in the folder",
            )
            return True

        return False

    def prepare(self, dest: str) -> None:
        """Prepare the context for a new folder processing."""
        self.dtstamp.refresh()
        self.folderstats.reset()
        self.quota.reset()
        self.reporter.reset(dest)

    def update_on_success(self, entry: FSEntry) -> None:
        """Update context on successful file operation."""
        self.folderstats.update(entry.size)
        self.quota.register_success(entry)

    def should_treat_as_dry_run(self, copy_path_str: str) -> bool:
        """Check if a file has already been transferred."""
        if self.is_dry_run:
            self.state = DryRunState(message=f"DRY - {copy_path_str}")
            return True
        return False

    def set_errored(self, copy_path_str: str) -> None:
        """Set the state to invalid file transfer."""
        self.state = TransferErrorState(message=f"FAILED - {copy_path_str}")

    def set_transferred(self, copy_path_str: str) -> None:
        """Set the state to successful file transfer."""
        self.state = TransferSuccessState(message=copy_path_str)

    def finalize(self, target: int, dest: str) -> str:
        """Finalize the context after processing."""
        none_found = self.is_none_found()
        report = self.reporter.generate_report(
            status=f"{self.state.status}: {self.folderstats.count}/{target} files copied",
            runtime=round(perf_counter() - self.folderstats.starttime, 2),
            size=self.folderstats.curr_size,
        )
        self.reporter.save()

        if none_found:
            remove_directory(dest)

        return report


@dataclass(slots=True)
class EngineState:
    """Abstract base class for engine states."""

    status: str = ""
    message: str = ""
    _context: Context | None = None

    def __post_init__(self) -> None:
        """Post-initialization tasks."""
        logger.debug("Change state to: %s", self.__class__.__name__)

    @property
    def context(self) -> Context:
        """Get the engine state context."""
        if self._context is None:
            msg = "Engine state context is not set."
            raise ValueError(msg)
        return self._context

    @context.setter
    def context(self, new_context: Context) -> None:
        """Set a new engine state context."""
        self._context = new_context


@dataclass(slots=True)
class RunningState(EngineState):
    """State representing engine running."""


@dataclass(slots=True)
class DryRunState(RunningState):
    """State representing engine running in dry-run mode."""


@dataclass(slots=True)
class TransferSuccessState(RunningState):
    """State representing successful file transfer."""


@dataclass(slots=True)
class TransferErrorState(RunningState):
    """State representing failed file transfer."""


@dataclass(slots=True)
class StoppedState(EngineState):
    """State representing engine stopped due to an error."""


@dataclass(slots=True)
class SuccessState(StoppedState):
    """State representing successful completion of folder processing."""


@dataclass(slots=True)
class UserStoppedState(StoppedState):
    """State representing user-requested stop of processing."""


@dataclass(slots=True)
class NoFilesFoundAllSearchedState(StoppedState):
    """State representing no files found in the folder and all files searched."""


@dataclass(slots=True)
class NoFilesFoundState(StoppedState):
    """State representing no files found in the folder."""


@dataclass(slots=True)
class AllSearched(StoppedState):
    """State representing all folders being searched."""


@dataclass(slots=True)
class FolderSizeLimitState(StoppedState):
    """State representing folder size limit reached."""


@dataclass(slots=True)
class TotalSizeLimitState(StoppedState):
    """State representing total size limit reached."""



"""Pydantic schemas for configuration."""

import os

from pydantic import BaseModel, field_validator

from ..utils import TransferMode


class FilecountModel(BaseModel):
    """Model for file count configuration."""

    count: int = 0
    is_rand_enabled: bool = False
    rand_min: int = 0
    rand_max: int = 0


class DirectoryModel(BaseModel):
    """Model for directory creation configuration."""

    is_enabled: bool = False
    is_unique: bool = True
    name: str = ""
    count: int = 1


class FilenameModel(BaseModel):
    """Model for file renaming."""

    template: str = "{original}"


class TransferModeModel(BaseModel):
    """Model for mode configuration."""

    transfer_mode: str = TransferMode.SYMLINK


class ListIncludeExcludeModel(BaseModel):
    """Model for list filtering."""

    is_enabled: bool = True
    should_include: bool = True
    text: str = ""


class MinMaxModel(BaseModel):
    """Model for size filter."""

    is_enabled: bool = False
    minimum: float = 0.0
    maximum: float = 0.0
    unit: str = ""


class SizeLimitModel(BaseModel):
    """Model for output folder size limits."""

    is_enabled: bool = False
    size_limit: float = 0.0
    unit: str = ""


class OptionsModel(BaseModel):
    """Model for additional options."""

    max_per_folder: int = 0
    should_follow_symlink: bool = False
    is_dry_run: bool = True


class ConfigModel(BaseModel):
    """Model for  configuration."""

    root: str
    dest: str
    filecount: FilecountModel
    folder: DirectoryModel
    filename: FilenameModel
    transfermode: TransferModeModel
    keyword: ListIncludeExcludeModel
    extension: ListIncludeExcludeModel
    filesize: MinMaxModel
    duration: MinMaxModel
    folder_size_limit: SizeLimitModel
    total_size_limit: SizeLimitModel
    options: OptionsModel

    @field_validator("root", "dest")
    @classmethod
    def is_absolute(cls, val: str) -> str:
        """Ensure root and dest paths are absolute."""
        if not os.path.isabs(val):
            return os.path.realpath(val)
        return val



"""CLI observer."""

import logging

from ..utils import Observer

logger = logging.getLogger(__name__)


class ConsoleObserver(Observer):
    """A simple console observer for fspachinko."""

    def on_progress_total(self, maximum: int) -> None:
        """Handle starting total progress."""
        logger.info("Starting total progress: %d directories(s)", maximum)

    def on_count_total(self) -> None:
        """Handle total progress count update."""
        logger.info("Total progress updated.")

    def on_progress(self, maximum: int) -> None:
        """Handle starting directory progress."""
        logger.info("Starting directory: %d file(s)", maximum)

    def on_finished(self) -> None:
        """Handle finishing process."""
        logger.info("Processing finished.")

    def on_log(self, msg: str) -> None:
        """Handle log message."""
        logger.info("%s", msg)

    def on_time(self) -> None:
        """Handle time update."""

    def on_count(self, count: int) -> None:
        """Handle directory progress count update."""



"""Core package."""

from .builder import build_engine
from .engine import Engine
from .quota import DiversityQuota
from .reporter import ReportWriter
from .transfer import fetch_transfer_strategy, get_available_transfer_modes
from .validator import FileValidator
from .walker import FSWalker

__all__ = [
    "DiversityQuota",
    "Engine",
    "FSWalker",
    "FileValidator",
    "ReportWriter",
    "build_engine",
    "fetch_transfer_strategy",
    "get_available_transfer_modes",
]



"""Helper functions for Qt GUI elements."""

from typing import TYPE_CHECKING, Any

from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QGroupBox,
    QLineEdit,
    QRadioButton,
    QSpinBox,
    QWidget,
)

from ..utils import strtobool

if TYPE_CHECKING:
    from collections.abc import Iterator

    from PySide6.QtCore import QObject
    from PySide6.QtGui import QAction


def get_qt_classname(obj: QObject) -> str:
    """Get the class name of a QObject."""
    return str(obj.metaObject().className())


def set_qt_name(w: QWidget | QAction, name: str) -> None:
    """Initialize a widget with a given object name."""
    w.setObjectName(name)


def set_qt_tips(w: QWidget | QAction, tooltip: str, statustip: str = "") -> None:
    """Set the tooltip and status tip for a widget."""
    if not statustip:
        statustip = f"{tooltip} | ({get_qt_classname(w)})"

    w.setToolTip(tooltip)
    w.setStatusTip(statustip)


def get_widget_value(widget: QWidget) -> Any:
    """Retrieve the value of a widget based on its type.

    Args:
        widget (QWidget): The widget to retrieve the value from.

    Returns:
        Any: The value of the widget, or None if not applicable.

    """
    match widget:
        case QLineEdit():
            return widget.text()
        case QComboBox():
            return widget.currentIndex()
        case QSpinBox() | QDoubleSpinBox():
            return widget.value()
        case QGroupBox() if not widget.isCheckable():
            return None
        case QCheckBox() | QRadioButton() | QGroupBox():
            return widget.isChecked()
        case _:
            return None


def set_widget_value(widget: QWidget, val: Any) -> None:
    """Set the value of a widget based on its type.

    Args:
        widget (QWidget): The widget to set the value for.
        val (Any): The value to set.

    """
    match widget:
        case QLineEdit():
            widget.setText(val)
        case QComboBox():
            try:
                index = int(val)
                if 0 <= index < widget.count():
                    widget.setCurrentIndex(index)
            except (ValueError, TypeError):
                pass
        case QSpinBox():
            widget.setValue(int(val))
        case QDoubleSpinBox():
            widget.setValue(float(val))
        case QCheckBox() | QRadioButton() | QGroupBox():
            state = strtobool(val=val)
            widget.setChecked(state)
        case _:
            return


def iter_custom_widget(w: QWidget) -> Iterator[tuple[str, QWidget]]:
    """Iterate over valid child widgets."""
    for child in w.findChildren(QWidget):
        if (key := child.objectName()) and not key.startswith("qt_"):
            yield key, child



"""Protocols for fspachinko."""

from abc import ABC, abstractmethod


class Observer(ABC):
    """Interface for fspachinko Observer."""

    @abstractmethod
    def on_progress_total(self, maximum: int) -> None:
        """Call when starting a new total progress cycle."""

    @abstractmethod
    def on_count_total(self) -> None:
        """Call to update total progress percentage."""

    @abstractmethod
    def on_progress(self, maximum: int) -> None:
        """Call when starting a new progress cycle."""

    @abstractmethod
    def on_finished(self) -> None:
        """Call when processing is finished."""

    @abstractmethod
    def on_log(self, msg: str) -> None:
        """Call to log a message."""

    @abstractmethod
    def on_time(self) -> None:
        """Call to update time remaining."""

    @abstractmethod
    def on_count(self, count: int) -> None:
        """Call to update progress percentage."""



"""Main entry point for GUI."""

import logging

from PySide6.QtCore import QCoreApplication
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication
from qt_material import apply_stylesheet

from ..utils import AppSetting, IconFilename, Paths, initialize_logging
from .mainwindow import MainWindow

logger = logging.getLogger(__name__)

QCoreApplication.setOrganizationName(AppSetting.ORGANIZATION)
QCoreApplication.setOrganizationDomain(AppSetting.DOMAIN)
QCoreApplication.setApplicationName(AppSetting.APPLICATION)


try:  # Windows only for taskbar icon
    from ctypes import windll

    myappid = "wonyoungjang.fspachinko.random_file_transfer_utility.0.0.1"
    windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
except ImportError:
    pass


def main() -> None:
    """Run the application."""
    initialize_logging()
    logger.info("Start: fspachinko GUI")

    app = QApplication()
    app.setWindowIcon(QIcon(Paths.icon(IconFilename.WINDOW)))
    apply_stylesheet(app, theme="dark_purple.xml", extra={"density_scale": "-2"})
    w = MainWindow()
    w.show()
    app.exec()


if __name__ == "__main__":
    main()



"""Builder module for core functionality."""

from random import Random
from typing import TYPE_CHECKING

from ..config import Filecount, Filename, Folder, ListIncludeExclude, MinMax, SizeLimit
from ..utils import SIZE_MAP, TIME_MAP, DateTimeStamp, ReStrFmt
from .engine import Engine
from .quota import DiversityQuota
from .reporter import ReportWriter
from .state import EngineContext
from .transfer import fetch_transfer_strategy
from .validator import FileValidator
from .walker import PachinkoFSWalker

if TYPE_CHECKING:
    from ..config import ConfigModel


def build_file_validator(m: ConfigModel) -> FileValidator:
    """Build and return a FileValidator based on the configuration."""
    keywords = ListIncludeExclude.from_model(m.keyword, re_fmt=ReStrFmt.KEYWORD)
    extensions = ListIncludeExclude.from_model(m.extension, re_fmt=ReStrFmt.EXTENSION)
    filesize = MinMax.from_model(m.filesize, mapping=SIZE_MAP)
    duration = MinMax.from_model(m.duration, mapping=TIME_MAP)
    return FileValidator(
        keywords=keywords,
        extensions=extensions,
        filesize=filesize,
        duration=duration,
    )


def build_engine(m: ConfigModel) -> Engine:
    """Build and return the engine based on the configuration."""
    # Build main components
    rng = Random()
    dtstamp = DateTimeStamp()

    # Build FileValidator
    validator = build_file_validator(m)

    # Build DiversityQuota
    quota = DiversityQuota(
        root=m.root,
        is_unique=m.folder.is_unique,
        max_per_dir=m.options.max_per_folder,
    )

    # Build EngineContext
    folder = Folder.from_model(m.folder, dest=m.dest)
    folder_size_limit = SizeLimit.from_model(m.folder_size_limit, mapping=SIZE_MAP)
    total_size_limit = SizeLimit.from_model(m.total_size_limit, mapping=SIZE_MAP)
    reporter = ReportWriter(
        root=m.root,
        exts_str=validator.extensions.as_string,
        keys_str=validator.keywords.as_string,
        dtstamp=dtstamp,
    )
    context = EngineContext(
        folder=folder,
        quota=quota,
        folder_size_limit=folder_size_limit,
        total_size_limit=total_size_limit,
        reporter=reporter,
        is_dry_run=m.options.is_dry_run,
        dtstamp=dtstamp,
    )

    # Build Walker
    walker = PachinkoFSWalker(
        root=m.root,
        quota=quota,
        validator=validator,
        rng=rng,
        should_follow_symlink=m.options.should_follow_symlink,
    )

    # Build Engine
    filecount = Filecount.from_model(m.filecount, rng=rng)
    filename = Filename.from_model(m.filename, dtstamp=dtstamp)
    do_transfer_strategy = fetch_transfer_strategy(m.transfermode.transfer_mode)
    return Engine(
        root=m.root,
        walker=walker,
        validator=validator,
        filecount=filecount,
        filename=filename,
        do_transfer_strategy=do_transfer_strategy,
        context=context,
        folder_count=m.folder.count,
    )



"""Config validation functions."""

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..utils import get_duration

if TYPE_CHECKING:
    import os
    from collections.abc import Callable

    from ..config import ListIncludeExclude, MinMax
    from .walker import FSEntry

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class FileValidator:
    """Class for validating files based on configuration."""

    keywords: ListIncludeExclude
    extensions: ListIncludeExclude
    filesize: MinMax
    duration: MinMax
    validators: tuple[Callable, ...] = ()

    def __post_init__(self) -> None:
        """Gather validation functions based on enabled filters."""
        self._gather_validators()

    def is_valid(self, entry: FSEntry) -> bool:
        """Check if a file is valid based on the current filters."""
        return all(is_valid(entry) for is_valid in self.validators)

    def _gather_validators(self) -> None:
        """Gather validation functions based on enabled filters."""
        v = []

        if self.filesize.is_enabled:
            v.append(self._is_valid_filesize)

        if self.extensions.is_enabled:
            v.append(self._is_valid_extension)

        if self.keywords.is_enabled:
            v.append(self._is_valid_keyword)

        self.validators = tuple(v)

    def _is_valid_filesize(self, entry: FSEntry) -> bool:
        """Check if a file is valid based on the current filters."""
        return self.filesize.is_valid(entry.size)

    def _is_valid_extension(self, entry: FSEntry) -> bool:
        """Check if a file is valid based on the current filters."""
        return self.extensions.is_valid(entry.ext)

    def _is_valid_keyword(self, entry: FSEntry) -> bool:
        """Check if a file is valid based on the current filters."""
        return self.keywords.is_valid(entry.stem)

    def is_valid_duration(self, entry: os.PathLike) -> bool:
        """Check if a file is valid based on the current filters."""
        if not self.duration.is_enabled:
            return True

        duration = get_duration(entry)
        return self.duration.is_valid(duration)



"""Workers for GUI."""

from dataclasses import dataclass
from typing import TYPE_CHECKING

from PySide6.QtCore import QThread, Slot

from ..core import build_engine
from .observer import GuiObserver, WorkerSignals

if TYPE_CHECKING:
    from ..config import ConfigModel
    from ..core import Engine


@dataclass(slots=True)
class MainWorker:
    """Worker for running process."""

    signals: WorkerSignals
    engine: Engine

    @classmethod
    def from_config(cls, config: ConfigModel) -> MainWorker:
        """Post-initialization tasks."""
        signals = WorkerSignals()
        observer = GuiObserver(signals)
        engine = build_engine(config)
        engine.set_observer(observer)
        return cls(signals, engine)

    def run(self) -> None:
        """Run the process."""
        self.engine.start()

    def stop(self) -> None:
        """Stop the process."""
        self.engine.request_stop()


class MainThread(QThread):
    """Worker thread for running process."""

    def __init__(self, worker: MainWorker) -> None:
        """Initialize the worker thread."""
        super().__init__()
        self.worker = worker

    @Slot()
    def run(self) -> None:
        """Run the process."""
        self.worker.run()

    def stop(self) -> None:
        """Stop the process."""
        self.worker.stop()



"""Convert between GUI profiles and configuration files."""

from collections.abc import Sequence
from typing import Any

from ..utils import ByteUnit, TimeUnit, TransferMode, load_json, save_json
from .schemas import ConfigModel


def _combo_value(data: dict[str, Any], key: str, default: str = "") -> str:
    val = data.get(key)
    if isinstance(val, str):
        return val

    items = data.get(f"{key}_items")
    if isinstance(val, int) and isinstance(items, Sequence) and 0 <= val < len(items):
        return str(items[val])

    return default


def _set_combo_value(profile: dict[str, Any], key: str, value: str, items: Sequence[str]) -> None:
    items_list = [str(i) for i in items]
    if value and value not in items_list:
        items_list.append(value)

    profile[f"{key}_items"] = items_list
    profile[key] = items_list.index(value) if value in items_list else 0


def convert_profile_to_config(profile_data: dict[str, Any]) -> dict[str, Any]:
    """Convert GUI profile data to configuration data."""
    root = _combo_value(profile_data, "root_combo")
    dest = _combo_value(profile_data, "dest_combo")

    should_create = bool(profile_data.get("folder", False))
    folder_count = int(profile_data.get("folder_count", 1)) if should_create else 1

    data = {
        "root": root,
        "dest": dest,
        "filecount": {
            "count": int(profile_data.get("filecount_fixed_val", 0)),
            "is_rand_enabled": bool(profile_data.get("filecount_rand_chk", False)),
            "rand_min": int(profile_data.get("filecount_rand_min", 0)),
            "rand_max": int(profile_data.get("filecount_rand_max", 0)),
        },
        "folder": {
            "should_create": should_create,
            "is_unique": bool(profile_data.get("folder_unique", True)),
            "name": str(profile_data.get("folder_name", "")),
            "count": folder_count,
        },
        "filename": {
            "template": str(profile_data.get("filename_template", "{original}")) or "{original}",
        },
        "transfermode": {
            "transfer_mode": _combo_value(profile_data, "transfermode_mode", TransferMode.SYMLINK),
        },
        "keyword": {
            "is_enabled": bool(profile_data.get("keyword", False)),
            "should_include": bool(profile_data.get("keyword_include", True)),
            "text": str(profile_data.get("keyword_text", "")),
        },
        "extension": {
            "is_enabled": bool(profile_data.get("extension", False)),
            "should_include": bool(profile_data.get("extension_include", True)),
            "text": str(profile_data.get("extension_text", "")),
        },
        "filesize": {
            "is_enabled": bool(profile_data.get("filesize", False)),
            "minimum": float(profile_data.get("filesize_minimum", 0.0)),
            "maximum": float(profile_data.get("filesize_maximum", 0.0)),
            "unit": _combo_value(profile_data, "filesize_unit"),
        },
        "duration": {
            "is_enabled": bool(profile_data.get("duration", False)),
            "minimum": float(profile_data.get("duration_minimum", 0.0)),
            "maximum": float(profile_data.get("duration_maximum", 0.0)),
            "unit": _combo_value(profile_data, "duration_unit"),
        },
        "folder_size_limit": {
            "is_enabled": bool(profile_data.get("folder_size_limit", False)),
            "size_limit": float(profile_data.get("folder_size_limit_size", 0.0)),
            "unit": _combo_value(profile_data, "folder_size_limit_unit"),
        },
        "total_size_limit": {
            "is_enabled": bool(profile_data.get("total_size_limit", False)),
            "size_limit": float(profile_data.get("total_size_limit_size", 0.0)),
            "unit": _combo_value(profile_data, "total_size_limit_unit"),
        },
        "options": {
            "max_per_folder": int(profile_data.get("options_max_per_folder", 0)),
            "should_follow_symlink": bool(profile_data.get("options_should_follow_symlink", False)),
            "is_dry_run": bool(profile_data.get("options_dry_run", False)),
        },
    }

    config = ConfigModel.model_validate(data)
    return config.model_dump()


def convert_config_to_profile(config_data: dict[str, Any]) -> dict[str, Any]:
    """Convert configuration data to GUI profile data."""
    config = ConfigModel.model_validate(config_data)
    data = config.model_dump()

    profile: dict[str, Any] = {
        "folder": data["folder"]["should_create"],
        "filecount_fixed_val": data["filecount"]["count"],
        "filecount_rand_chk": data["filecount"]["is_rand_enabled"],
        "filecount_rand_min": data["filecount"]["rand_min"],
        "filecount_rand_max": data["filecount"]["rand_max"],
        "folder_count": data["folder"]["count"],
        "folder_name": data["folder"]["name"],
        "folder_unique": data["folder"]["is_unique"],
        "filename_template": data["filename"]["template"],
        "keyword": data["keyword"]["is_enabled"],
        "keyword_include": data["keyword"]["should_include"],
        "keyword_text": data["keyword"]["text"],
        "extension": data["extension"]["is_enabled"],
        "extension_include": data["extension"]["should_include"],
        "extension_text": data["extension"]["text"],
        "filesize": data["filesize"]["is_enabled"],
        "filesize_minimum": data["filesize"]["minimum"],
        "filesize_maximum": data["filesize"]["maximum"],
        "duration": data["duration"]["is_enabled"],
        "duration_minimum": data["duration"]["minimum"],
        "duration_maximum": data["duration"]["maximum"],
        "folder_size_limit": data["folder_size_limit"]["is_enabled"],
        "folder_size_limit_size": data["folder_size_limit"]["size_limit"],
        "total_size_limit": data["total_size_limit"]["is_enabled"],
        "total_size_limit_size": data["total_size_limit"]["size_limit"],
        "options_max_per_folder": data["options"]["max_per_folder"],
        "options_should_follow_symlink": data["options"]["should_follow_symlink"],
        "options_dry_run": data["options"]["is_dry_run"],
    }

    _set_combo_value(profile, "root_combo", data["root"], [data["root"]])
    _set_combo_value(profile, "dest_combo", data["dest"], [data["dest"]])
    _set_combo_value(profile, "filesize_unit", data["filesize"]["unit"], list(ByteUnit))
    _set_combo_value(profile, "duration_unit", data["duration"]["unit"], list(TimeUnit))
    _set_combo_value(profile, "folder_size_limit_unit", data["folder_size_limit"]["unit"], list(ByteUnit))
    _set_combo_value(profile, "total_size_limit_unit", data["total_size_limit"]["unit"], list(ByteUnit))

    transfer_modes = list(TransferMode)
    _set_combo_value(profile, "transfermode_mode", data["transfermode"]["transfer_mode"], transfer_modes)

    return profile


def profile_to_config_file(profile_path: str, config_path: str) -> None:
    """Convert GUI profile JSON to configuration JSON."""
    profile_data = load_json(profile_path)
    config_data = convert_profile_to_config(profile_data)
    save_json(config_path, config_data)


def config_to_profile_file(config_path: str, profile_path: str) -> None:
    """Convert configuration JSON to GUI profile JSON."""
    config_data = load_json(config_path)
    profile_data = convert_config_to_profile(config_data)
    save_json(profile_path, profile_data)



"""Settings handling for GUI."""

import os
from collections.abc import Sequence
from dataclasses import dataclass, field

from PySide6.QtWidgets import QComboBox, QWidget

from ..utils import Paths, load_json, save_json
from .qthelpers import get_widget_value, iter_custom_widget, set_widget_value


@dataclass(slots=True)
class ProfileManager:
    """Class for managing GUI profiles."""

    current_profile: str = field(init=False)

    def set_current(self, profile: str) -> None:
        """Set the current profile name."""
        self.current_profile = Paths.profile(profile)

    def save_profile(self, parent: QWidget) -> None:
        """Recursively save settings for all child widgets."""
        data = {}
        for key, child in iter_custom_widget(parent):
            if (val := get_widget_value(child)) is None:
                continue

            data[key] = val
            if isinstance(child, QComboBox):
                items = [child.itemText(i) for i in range(child.count())]
                data[f"{key}_items"] = items

        save_json(self.current_profile, data)

    def open_profile(self, parent: QWidget) -> None:
        """Recursively load settings for all child widgets."""
        data = load_json(self.current_profile)

        for key, child in iter_custom_widget(parent):
            if isinstance(child, QComboBox):
                items = data.get(f"{key}_items")
                if isinstance(items, Sequence):
                    child.clear()
                    child.addItems([str(i) for i in items])

            if (val := data.get(key)) is not None:
                set_widget_value(child, val)

    def get_current_profile_parent(self) -> str:
        """Get the parent directory of the current profile."""
        return os.path.dirname(self.current_profile)



"""Actions module for QActions."""

from dataclasses import dataclass, field

from PySide6.QtGui import QAction, QIcon

from ..utils import IconFilename, Paths
from .components import set_qt_tips


@dataclass(slots=True)
class FileActions:
    """Main file menu actions."""

    save: QAction = field(init=False)
    save_as: QAction = field(init=False)
    load: QAction = field(init=False)
    autosave: QAction = field(init=False)
    exit: QAction = field(init=False)

    def __post_init__(self) -> None:
        """Initialize the main actions."""
        self.save = QAction(icon=QIcon(Paths.icon(IconFilename.SAVE)), text="&Save Profile")
        self.save.setShortcut("Ctrl+S")
        set_qt_tips(self.save, "Save current GUI profile (Ctrl+S)")

        self.save_as = QAction(icon=QIcon(Paths.icon(IconFilename.SAVE_AS)), text="Save Profile &As")
        self.save_as.setShortcut("Ctrl+Shift+S")
        set_qt_tips(self.save_as, "Save current GUI profile as ... (Ctrl+Shift+S)")

        self.load = QAction(icon=QIcon(Paths.icon(IconFilename.OPEN)), text="&Load Profile")
        self.load.setShortcut("Ctrl+O")
        set_qt_tips(self.load, "Load GUI profile (Ctrl+O)")

        self.autosave = QAction(
            icon=QIcon(Paths.icon(IconFilename.AUTOSAVE)), text="A&utosave Profile", checkable=True, checked=True
        )
        set_qt_tips(self.autosave, "Automatically save profile on exit")

        self.exit = QAction(icon=QIcon(Paths.icon(IconFilename.CLOSE)), text="&Exit")
        self.exit.setShortcut("Ctrl+W")
        set_qt_tips(self.exit, "Exit application (Ctrl+W)")


@dataclass(slots=True)
class RunActions:
    """Main run actions."""

    start: QAction = field(init=False)
    stop: QAction = field(init=False)

    def __post_init__(self) -> None:
        """Initialize the main actions."""
        self.start = QAction(icon=QIcon(Paths.icon(IconFilename.START)), text="&Start")
        self.start.setShortcut("Ctrl+R")
        set_qt_tips(self.start, "Start (Ctrl+R)")

        self.stop = QAction(icon=QIcon(Paths.icon(IconFilename.STOP)), text="S&top")
        self.stop.setShortcut("ESC")
        set_qt_tips(self.stop, "Stop (ESC)")


@dataclass(slots=True)
class Actions:
    """Main actions."""

    file: FileActions = field(default_factory=FileActions)
    run: RunActions = field(default_factory=RunActions)



"""File transfer strategies."""

import os
import shutil
from io import UnsupportedOperation
from os import PathLike
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

from ..utils import FileError, TransferMode

if TYPE_CHECKING:
    from collections.abc import Callable


def get_available_transfer_modes() -> tuple[TransferMode, ...]:
    """Detect which transfer modes are supported on the current OS."""
    # COPY and MOVE are always available
    available = [TransferMode.COPY, TransferMode.COPY_PRESERVE, TransferMode.MOVE]

    # SYMLINK availability
    try:
        with TemporaryDirectory() as tmpdir:
            test_src = os.path.join(tmpdir, "test_src")
            test_dst = os.path.join(tmpdir, "test_symlink")
            open(test_src, "w").close()
            os.symlink(test_src, test_dst)
            os.unlink(test_dst)
            os.unlink(test_src)
        available.append(TransferMode.SYMLINK)
    except (OSError, UnsupportedOperation, NotImplementedError):
        pass

    # HARDLINK availability
    try:
        with TemporaryDirectory() as tmpdir:
            test_src = os.path.join(tmpdir, "test_src")
            test_dst = os.path.join(tmpdir, "test_hardlink")
            open(test_src, "w").close()
            os.link(test_src, test_dst)
            os.unlink(test_dst)
            os.unlink(test_src)
        available.append(TransferMode.HARDLINK)
    except (OSError, UnsupportedOperation, NotImplementedError):
        pass

    return tuple(available)


def fetch_transfer_strategy(mode: str) -> Callable[[PathLike, str], None]:
    """Return the appropriate transfer strategy instance.

    Falls back to SYMLINK if the requested mode is not available.
    """
    strategy_map = {
        TransferMode.COPY: transfer_copy,
        TransferMode.COPY_PRESERVE: transfer_copy_preserve,
        TransferMode.MOVE: transfer_move,
        TransferMode.SYMLINK: transfer_symlink,
        TransferMode.HARDLINK: transfer_hardlink,
    }
    available_modes = get_available_transfer_modes()
    requested_mode = TransferMode(mode)
    if requested_mode in available_modes:
        return strategy_map[requested_mode]
    return transfer_symlink


def transfer_copy(src: PathLike, dst: str) -> None:
    """Copy a file from source to destination."""
    shutil.copy(src, dst)


def transfer_copy_preserve(src: PathLike, dst: str) -> None:
    """Copy a file from source to destination preserving metadata."""
    shutil.copy2(src, dst)


def transfer_move(src: PathLike, dst: str) -> None:
    """Move a file from source to destination."""
    shutil.move(src, dst)


def transfer_symlink(src: PathLike, dst: str) -> None:
    """Create a symlink from source to destination."""
    os.symlink(src, dst)


def transfer_hardlink(src: PathLike, dst: str) -> None:
    """Create a hardlink from source to destination.

    Falls back to symlink if hardlinking across filesystems fails.
    """
    try:
        os.link(src, dst)
    except OSError as e:
        if e.winerror == FileError.WINDOWS_CROSS_DRIVE_ERROR or e.errno == FileError.UNIX_CROSS_FILESYSTEM_ERROR:
            os.symlink(src, dst)
        else:
            raise



"""Engine Module."""

import logging
import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from ..config import Filecount, Filename
    from ..utils import Observer
    from .state import Context
    from .validator import FileValidator
    from .walker import FSWalker


logger = logging.getLogger(__name__)


@dataclass(slots=True)
class Engine:
    """Core engine class."""

    root: str
    walker: FSWalker
    validator: FileValidator
    filecount: Filecount
    filename: Filename
    do_transfer_strategy: Callable[[os.PathLike, str], None]
    context: Context
    folder_count: int
    observer: Observer = field(init=False)

    def set_observer(self, observer: Observer) -> None:
        """Set the observer for the engine."""
        self.observer = observer

    def request_stop(self) -> None:
        """Request to stop the engine."""
        self.context.is_stop_requested = True

    def start(self) -> None:
        """Run the main file copying process."""
        self.observer.on_progress_total(self.folder_count)
        for target, dest in self.get_transfer_parameters():
            if not self.context.quota.is_unique:
                self.walker.reset()
            self.process_directory(target, dest)
        self.observer.on_finished()

    def get_transfer_parameters(self) -> Iterator[tuple[int, str]]:
        """Get transfer parameters for all folders."""
        for _ in range(self.folder_count):
            yield (
                self.filecount.get_file_count(),
                self.context.folder.determine_dest_dirname(),
            )

    def process_directory(self, target: int, dest: str) -> None:
        """Run processing for a single folder."""
        os.mkdir(dest)
        self.observer.on_progress(target)
        self.context.prepare(dest)
        self.transfer_directory(target, dest)
        self.observer.on_count_total()
        self.report(self.context.finalize(target, dest))

    def transfer_directory(self, target: int, dest: str) -> None:
        """Process a single folder for file copying."""
        if self.context.should_stop(target):
            self.report_state()
            return

        for entry in self.walker.walk():
            if entry is None:
                break

            if self.context.should_stop(target):
                self.report_state()
                return

            if not self.validator.is_valid_duration(entry):
                continue

            if not self.transfer_file(entry, dest):
                continue

            self.context.update_on_success(entry)
            self.update_observer_on_entry()

    def transfer_file(self, entry: os.PathLike, dest: str) -> bool:
        """Attempt to copy a file and return success status."""
        count = self.context.folderstats.count
        chosen_rel = os.path.relpath(entry, self.root)
        chosen_new = self.filename.determine_dest_filename(chosen_rel, dest, count)
        if chosen_new is None:
            return False

        msg = f"{count + 1}: {chosen_rel} -> {os.path.relpath(chosen_new, dest)}"

        if self.context.should_treat_as_dry_run(msg):
            self.report_state()
            return True

        try:
            self.do_transfer_strategy(entry, chosen_new)
        except (PermissionError, OSError):
            self.context.set_errored(msg)
            self.report_state()
            return False

        self.context.set_transferred(msg)
        self.report_state()
        return True

    def report_state(self) -> None:
        """Report the current engine state."""
        self.report(msg=self.context.state.message)

    def report(self, msg: str) -> None:
        """Report and log a message."""
        self.observer.on_log(msg)
        self.context.reporter.record(msg)

    def update_observer_on_entry(self) -> None:
        """Update observer with current entry statistics."""
        self.observer.on_count(self.context.folderstats.count)
        self.observer.on_time()



"""Reporter for process."""

import os
from dataclasses import dataclass, field

from ..utils import DateTimeStamp, convert_byte_to_human_readable_size


@dataclass(slots=True)
class ReportWriter:
    """ReportWriter class."""

    root: str
    exts_str: str
    keys_str: str
    dtstamp: DateTimeStamp
    buffer: list[str] = field(default_factory=list)
    report_path: str = field(init=False)
    dest: str = field(init=False)

    def reset(self, dest: str) -> None:
        """Initialize reporter for a new run."""
        self.buffer.clear()
        self.report_path = os.path.join(dest, f"!_report_{os.path.basename(dest)}.txt")
        self.dest = dest

    def record(self, message: str) -> None:
        """Add a message to the buffer."""
        self.buffer.append(f"{message}\n")

    def generate_report(self, status: str, runtime: float, size: int) -> str:
        """Generate the header report string."""
        return (
            f"\n{status}"
            "\n------------------------------------------------------------------------\n"
            f"Date:             {self.dtstamp.date_time_report_str}\n"
            f"Root:             {self.root}\n"
            f"Destination:      {self.dest}\n"
            f"Extensions:       {self.exts_str}\n"
            f"Keywords:         {self.keys_str}\n"
            f"Total size:       {convert_byte_to_human_readable_size(size)}\n"
            f"Total runtime:    {runtime}s\n"
            "\n========================================================================\n"
            "========================================================================\n"
            "========================================================================\n"
            "========================================================================\n"
            "========================================================================\n"
        )

    def save(self) -> None:
        """Save the report to file."""
        mode = "a" if os.path.exists(self.report_path) else "w"
        with open(self.report_path, mode=mode, encoding="utf-8") as f:
            f.writelines(self.buffer)
            f.write("\n\n")



"""Main module."""

import logging

from PySide6.QtCore import Slot
from PySide6.QtWidgets import QGroupBox, QWidget

from ..utils import PERCENTAGE_100, GUIName
from .components import ProgressBinder
from .qthelpers import set_qt_name
from .uibuilder import UIBuilder
from .workers import MainThread, MainWorker

logger = logging.getLogger(__name__)


class CentralWidget(QWidget):
    """Main application window."""

    def __init__(self) -> None:
        """Initialize the main window."""
        super().__init__()
        set_qt_name(self, GUIName.CENTRAL_WIDGET)
        self.main_thread: MainThread | None = None
        self.ui = UIBuilder()
        layout = self.ui.build_layout()
        self.setLayout(layout)
        self.progress_binder = ProgressBinder(self.ui.progress, self.ui.logging)
        self.window_title_original = ""

    @Slot()
    def on_start(self) -> None:
        """Start the process and disable UI elements."""
        try:
            config = self.ui.get_config()
        except Exception:
            logger.exception("")
            return

        self.main_thread = MainThread(MainWorker.from_config(config))
        self.ui.progress.reset()
        self.window_title_original = self.window().windowTitle()
        self.progress_binder.bind(self.main_thread.worker.signals)
        self.progress_binder.count.connect(self.update_title_progress)
        self.progress_binder.finished.connect(self.on_finished)
        self.toggle_ui(is_enabled=False)
        self.main_thread.start()

    @Slot()
    def on_stop(self) -> None:
        """Stop the process."""
        if self.main_thread and self.main_thread.isRunning():
            self.main_thread.stop()

    @Slot(int)
    def update_title_progress(self, val: int) -> None:
        """Update window title with progress percentage."""
        pct = int((val / self.ui.progress.progbar_dir.maximum()) * PERCENTAGE_100)
        self.window().setWindowTitle(f"[{pct}%] {self.window_title_original}")

    @Slot()
    def on_finished(self) -> None:
        """Handle worker finished signal."""
        self.toggle_ui(is_enabled=True)
        self.window().setWindowTitle(self.window_title_original)

    def toggle_ui(self, *, is_enabled: bool) -> None:
        """Lock or unlock UI elements."""
        for child in self.findChildren(QGroupBox):
            child.setEnabled(is_enabled)



"""Quota and State management."""

import logging
import os
from collections import Counter
from dataclasses import dataclass, field

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class DiversityQuota:
    """Manages rules for diversity (weights) and uniqueness."""

    root: str
    is_unique: bool
    max_per_dir: int
    locked_dir: set[str] = field(default_factory=set)
    dircount: Counter[str] = field(default_factory=Counter)

    def reset(self) -> None:
        """Reset batch-specific counters, optionally keeping file history."""
        self.dircount.clear()
        self.locked_dir.clear()

    def is_all_locked(self) -> bool:
        """Check if all files/folders are locked."""
        return self.root in self.locked_dir

    def is_dir_locked(self, directory: str) -> bool:
        """Check if a folder is locked."""
        return directory in self.locked_dir

    def lock_dir(self, directory: str) -> None:
        """Mark a folder as locked without registering a success."""
        self.locked_dir.add(directory)

    def register_success(self, entry: os.PathLike) -> None:
        """Record a successful copy and apply locking rules."""
        if self.max_per_dir <= 0:
            return

        leaf_dir = os.path.dirname(entry)
        self.dircount[leaf_dir] += 1
        if self.dircount[leaf_dir] >= self.max_per_dir:
            self.lock_dir(leaf_dir)



"""Utility functions."""

import contextlib
import json
import logging
import os
import shutil
import subprocess
from filecmp import cmp
from typing import Any

from .constants import DURATION_CMD, FALSE_STRS, TRUE_STRS, BytesIn, ByteUnit

logger = logging.getLogger(__name__)


class SafeDict(dict):
    """A helper class for string formatting.

    If a key is missing, it returns the key wrapped in braces
    instead of raising a KeyError.
    """

    def __missing__(self, key: str) -> str:
        """Return the key wrapped in braces if missing."""
        return "{" + key + "}"


def calc_unique_path_name(dest: str, stem_or_name: str, ext: str = "") -> str:
    """Calculate a unique path name in the destination."""
    target = os.path.join(dest, f"{stem_or_name}{ext}")

    x = 2
    while os.path.exists(target):
        target = os.path.join(dest, f"{stem_or_name} ({x}){ext}")
        x += 1

    return target


def strtobool(*, val: str | int | bool) -> bool:
    """Convert a string representation of truth to true (1) or false (0).

    Replaces distutils.util.strtobool function (deprecated in Python 3.10).

    True values are: y, yes, t, true, on, 1
    False values are: n, no, f, false, off, 0

    Raises ValueError if val is anything else.
    """
    if isinstance(val, bool):
        return val

    val_str = str(val).casefold()

    if val_str in TRUE_STRS:
        return True

    if val_str in FALSE_STRS:
        return False

    msg = f"Invalid truth value {val!r}"
    raise ValueError(msg)


def convert_string_to_list(string: str, sep: str = ",") -> tuple[str, ...]:
    """Convert a comma-separated string to a list."""
    if not string:
        return ()

    li = tuple(s.strip() for s in string.split(sep))
    if len(li) == 1 and li[0] == "":
        return ()
    return li


def convert_byte_to_human_readable_size(nbytes: int) -> str:
    """Convert bytes to human readable string."""
    if nbytes < BytesIn.KILOBYTE:
        return f"{nbytes} {ByteUnit.BYTES}"

    if nbytes < BytesIn.MEGABYTE:
        return f"{round(nbytes / BytesIn.KILOBYTE, 2)} {ByteUnit.KILOBYTES}"

    if nbytes < BytesIn.GIGABYTE:
        return f"{round(nbytes / BytesIn.MEGABYTE, 2)} {ByteUnit.MEGABYTES}"

    return f"{round(nbytes / BytesIn.GIGABYTE, 2)} {ByteUnit.GIGABYTES}"


def remove_directory(path: str) -> None:
    """Remove a directory and its contents."""
    with contextlib.suppress(OSError):
        shutil.rmtree(path)


def are_paths_equal(path1: str, path2: str) -> bool:
    """Compare two paths for equality, accounting for case sensitivity."""
    if cmp(path1, path2, shallow=True):
        return True
    return cmp(path1, path2, shallow=False)


def load_json(path: str) -> dict[str, Any]:
    """Load JSON data from a file and return as a dictionary."""
    if not (os.path.exists(path) and os.path.isfile(path)):
        return {}

    with open(path, encoding="utf-8") as f:
        return json.load(f)


def save_json(path: str, data: dict[str, Any]) -> None:
    """Save a dictionary as JSON data to a file."""
    os.makedirs(os.path.dirname(path) or ".", exist_ok=True)
    with open(path, "w", encoding="utf-8") as f:
        data = dict(sorted(data.items(), key=lambda item: item[0]))
        json.dump(data, f, indent=4)


def get_stem_and_ext(path: str) -> tuple[str, str]:
    """Get the stem and extension of a file path."""
    return os.path.splitext(os.path.basename(path))


def get_duration(path: os.PathLike) -> float:
    """Get the duration of a media file."""
    try:
        out_bytes = subprocess.check_output(
            [*DURATION_CMD, path],
            stderr=subprocess.DEVNULL,
            timeout=10,
        )
        try:
            return float(out_bytes.decode().strip())
        except ValueError:
            logger.debug("ffprobe output could not be parsed as float: %s", out_bytes.decode(errors="ignore"))
            return 0.0
    except subprocess.CalledProcessError as e:
        out_bytes = e.output
        code = e.returncode
        logger.debug("ffprobe failed with code %d: %s", code, out_bytes.decode(errors="ignore"))
        return 0.0



"""Logging configuration."""

import logging.config

from .constants import DefaultPath
from .helpers import load_json
from .paths import Paths


def initialize_logging(path: str | None = None) -> None:
    """Initialize logging for the application."""
    if path is None:
        path = Paths.config(DefaultPath.LOGGING)

    data = load_json(path)
    logging.config.dictConfig(data)



"""Configuration dataclasses."""

import os
import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ..utils import (
    INVALID_FILENAME_CHARS,
    DateTimeStamp,
    FilenameTemplateMapKey,
    SafeDict,
    are_paths_equal,
    calc_unique_path_name,
    convert_string_to_list,
    get_stem_and_ext,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from random import Random

    from .schemas import (
        DirectoryModel,
        FilecountModel,
        FilenameModel,
        ListIncludeExcludeModel,
        MinMaxModel,
        SizeLimitModel,
    )


@dataclass(slots=True)
class Filecount:
    """Dataclass for file count configuration."""

    count: int
    is_rand_enabled: bool
    rand_min: int
    rand_max: int
    rng: Random

    @classmethod
    def from_model(cls, m: FilecountModel, rng: Random) -> Filecount:
        """Create Filecount from configuration model."""
        return cls(
            count=m.count,
            is_rand_enabled=m.is_rand_enabled,
            rand_min=m.rand_min,
            rand_max=m.rand_max,
            rng=rng,
        )

    def get_file_count(self) -> int:
        """Get the file count based on configuration."""
        if self.is_rand_enabled:
            return self.rng.randint(self.rand_min, self.rand_max)
        return self.count


@dataclass(slots=True)
class Filename:
    """Dataclass for file renaming."""

    template: str
    dtstamp: DateTimeStamp

    @classmethod
    def from_model(cls, m: FilenameModel, dtstamp: DateTimeStamp) -> Filename:
        """Create Filename from configuration model."""
        return cls(template=m.template, dtstamp=dtstamp)

    def calc_target_name(self, chosen: str, dest: str, index: int) -> str:
        """Prepare the target file path based on naming conventions."""
        stem, ext = get_stem_and_ext(chosen)

        safe_dict = SafeDict(
            {
                FilenameTemplateMapKey.DATE: self.dtstamp.date,
                FilenameTemplateMapKey.TIME: self.dtstamp.time,
                FilenameTemplateMapKey.DATETIME: self.dtstamp.date_time,
                FilenameTemplateMapKey.ORIGINAL: stem,
                FilenameTemplateMapKey.INDEX: index + 1,
                FilenameTemplateMapKey.PARENT: os.path.basename(os.path.dirname(chosen)),
                FilenameTemplateMapKey.PARENTS_TO_ROOT: "_".join(chosen.split(os.sep)[:-1]),
            }
        )

        try:
            new_stem = self.template.format_map(safe_dict)
        except (KeyError, ValueError):
            new_stem = stem

        name = "".join(c for c in new_stem if c not in INVALID_FILENAME_CHARS) + ext
        return os.path.join(dest, name)

    def determine_dest_filename(self, chosen: str, dest: str, index: int) -> str | None:
        """Calculate the destination file path based on configuration."""
        target = self.calc_target_name(chosen, dest, index)

        if not os.path.exists(target):
            return target

        if are_paths_equal(chosen, target):
            return None

        stem, ext = get_stem_and_ext(target)
        return calc_unique_path_name(dest, stem, ext)


@dataclass(slots=True)
class Folder:
    """Dataclass for folder creation configuration."""

    is_enabled: bool
    dest: str
    name: str

    @classmethod
    def from_model(cls, m: DirectoryModel, dest: str) -> Folder:
        """Create Folder from configuration model."""
        return cls(
            is_enabled=m.is_enabled,
            dest=dest,
            name=m.name,
        )

    def determine_dest_dirname(self) -> str:
        """Calculate the destination directory name based on configuration."""
        if not self.is_enabled:
            return self.dest
        return calc_unique_path_name(self.dest, self.name)


@dataclass(slots=True)
class MinMax:
    """Dataclass for min-max limit configuration."""

    is_enabled: bool
    minimum: float
    maximum: float

    @classmethod
    def from_model(cls, m: MinMaxModel, mapping: dict[str, float]) -> MinMax:
        """Create MinMax from configuration model."""
        return cls(
            is_enabled=m.is_enabled,
            minimum=m.minimum * mapping.get(m.unit, 1.0),
            maximum=m.maximum * mapping.get(m.unit, 1.0),
        )

    def is_valid(self, value: float) -> bool:
        """Check if a value is within the min-max range."""
        return self.minimum <= value <= self.maximum


@dataclass(slots=True)
class SizeLimit:
    """Dataclass for output folder size limits."""

    is_enabled: bool
    size_limit: float

    @classmethod
    def from_model(cls, m: SizeLimitModel, mapping: dict[str, float]) -> SizeLimit:
        """Create SizeLimit from configuration model."""
        return cls(
            is_enabled=m.is_enabled,
            size_limit=m.size_limit * mapping.get(m.unit, 1.0),
        )

    def is_valid(self, size: int) -> bool:
        """Check if the size limit is exceeded."""
        return size > self.size_limit


@dataclass(slots=True)
class ListIncludeExclude:
    """Dataclass for include-exclude list configuration."""

    is_enabled: bool
    should_include: bool
    as_string: str
    patterns: tuple[re.Pattern, ...]
    is_valid: Callable[[str], bool] = field(init=False)

    def __post_init__(self) -> None:
        """Post init to set validator function."""
        self.is_valid = self._is_valid_include if self.should_include else self._is_valid_exclude

    @classmethod
    def from_model(cls, m: ListIncludeExcludeModel, re_fmt: str) -> ListIncludeExclude:
        """Create ListIncludeExclude from configuration model."""
        as_string = ""
        patterns = ()
        if text := m.text.strip():
            text_list = convert_string_to_list(text)
            as_string = ", ".join(text_list)
            patterns = tuple(re.compile(re_fmt.format(re.escape(i)), re.IGNORECASE) for i in text_list)
        return cls(
            is_enabled=m.is_enabled and bool(text),
            should_include=m.should_include,
            as_string=as_string,
            patterns=patterns,
        )

    def _is_valid_include(self, part: str) -> bool:
        """Check if a file name part matches the include regexes."""
        return any(p.search(part) for p in self.patterns)

    def _is_valid_exclude(self, part: str) -> bool:
        """Check if a file name part matches the exclude regexes."""
        return not any(p.search(part) for p in self.patterns)



"""GUI components in PySide6."""

import logging
import os
from typing import TYPE_CHECKING, ClassVar

from PySide6.QtCore import QDir, QObject, QUrl, Signal, Slot
from PySide6.QtGui import QDesktopServices, QIcon
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QFileDialog,
    QFormLayout,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLineEdit,
    QMenu,
    QProgressBar,
    QPushButton,
    QRadioButton,
    QSpinBox,
    QTextBrowser,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from ..config import (
    DirectoryModel,
    FilecountModel,
    FilenameModel,
    ListIncludeExcludeModel,
    MinMaxModel,
    OptionsModel,
    SizeLimitModel,
    TransferModeModel,
)
from ..core import get_available_transfer_modes
from ..utils import ByteUnit, FilenameTemplate, IconFilename, Paths, TimeUnit
from .qthelpers import set_qt_name, set_qt_tips

if TYPE_CHECKING:
    from collections.abc import Sequence

    from PySide6.QtGui import QDragEnterEvent, QDropEvent

    from .workers import WorkerSignals

logger = logging.getLogger(__name__)


class BaseGroupBox(QGroupBox):
    """Base class for group boxes with common functionality."""

    def __init__(self, title: str, name: str, *, checkable: bool = False) -> None:
        """Initialize the base group box."""
        super().__init__(title=title)
        set_qt_name(self, name)
        self.setCheckable(checkable)


class PathSelectorWidget(BaseGroupBox):
    """Handles logic for selecting a path."""

    def __init__(self, title: str, name: str, items: Sequence[str]) -> None:
        """Initialize the path selector widget."""
        super().__init__(title, name)
        self.setAcceptDrops(True)

        title_lower = self.title().casefold()

        self.combo = QComboBox()
        self.combo.addItems(items)
        set_qt_name(self.combo, f"{name}_combo")
        set_qt_tips(self.combo, f"Select or enter a path for {title_lower}.")

        icon_browse = QIcon(Paths.icon(IconFilename.BROWSE))
        icon_open = QIcon(Paths.icon(IconFilename.OPEN_DIR))
        icon_delete = QIcon(Paths.icon(IconFilename.REMOVE))

        self.btn_browse = QPushButton()
        self.btn_browse.setIcon(icon_browse)
        self.btn_browse.clicked.connect(self.browse)
        set_qt_name(self.btn_browse, f"{name}_browse_btn")
        set_qt_tips(self.btn_browse, f"Browse for {title_lower} folder.")

        self.btn_open = QPushButton()
        self.btn_open.setIcon(icon_open)
        self.btn_open.clicked.connect(self.open)
        set_qt_name(self.btn_open, f"{name}_open_btn")
        set_qt_tips(self.btn_open, f"Open current {title_lower} folder in file explorer.")

        self.btn_delete = QPushButton()
        self.btn_delete.setIcon(icon_delete)
        self.btn_delete.clicked.connect(self.delete_curr_item)
        set_qt_name(self.btn_delete, f"{name}_delete_btn")
        set_qt_tips(self.btn_delete, f"Delete current {title_lower} entry.")

        layout = QHBoxLayout(self)
        layout.addWidget(self.combo, stretch=1)
        layout.addWidget(self.btn_browse)
        layout.addWidget(self.btn_delete)
        layout.addWidget(self.btn_open)

    def dragEnterEvent(self, event: QDragEnterEvent) -> None:  # noqa: N802
        """Handle drag enter event for folder paths."""
        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent) -> None:  # noqa: N802
        """Handle drop event for folder paths."""
        for url in event.mimeData().urls():
            path = url.toLocalFile()
            if os.path.isdir(path):
                if self.combo.findText(path) == -1:
                    self.combo.addItem(path)
                self.combo.setCurrentText(path)

    @Slot()
    def browse(self) -> None:
        """Return the browse button."""
        d = QFileDialog.getExistingDirectory(self, f"Select {self.title()}")
        if d:
            if self.combo.findText(d) == -1:
                self.combo.addItem(d)
            self.combo.setCurrentText(d)

    @Slot()
    def delete_curr_item(self) -> None:
        """Delete the currently selected item."""
        if self.combo.count() > 0:
            self.combo.removeItem(self.combo.currentIndex())

    @Slot()
    def open(self) -> None:
        """Open the currently selected path in file explorer."""
        try:
            QDesktopServices.openUrl(QUrl.fromLocalFile(self.combo.currentText()))
        except Exception:
            logger.exception("Failed to open path")

    def get_config(self) -> str:
        """Return clean data for the config."""
        return self.combo.currentText()


class RootPathSelectorWidget(PathSelectorWidget):
    """Handles logic for selecting a root path."""

    def __init__(self) -> None:
        """Initialize the root path selector widget."""
        super().__init__("Root", "root", [QDir.rootPath()])


class DestPathSelectorWidget(PathSelectorWidget):
    """Handles logic for selecting a destination path."""

    def __init__(self) -> None:
        """Initialize the destination path selector widget."""
        super().__init__("Destination", "dest", [QDir.homePath()])


class FileCountWidget(BaseGroupBox):
    """Handles logic for file count settings."""

    def __init__(self) -> None:
        """Initialize the file count widget."""
        name = "filecount"
        super().__init__("File count", name)

        self.radio_fixed = QRadioButton("Fixed")
        self.radio_fixed.setChecked(True)
        set_qt_name(self.radio_fixed, f"{name}_fixed_chk")
        set_qt_tips(self.radio_fixed, "Select fixed file count.")

        self.spin_fixed = QSpinBox(suffix=" files")
        set_qt_name(self.spin_fixed, f"{name}_fixed_val")
        set_qt_tips(self.spin_fixed, "Number of files to copy.")

        self.radio_rand = QRadioButton("Random")
        set_qt_name(self.radio_rand, f"{name}_rand_chk")
        set_qt_tips(self.radio_rand, "Select random file count.")

        self.spin_min_rand = QSpinBox(prefix="Min: ")
        set_qt_name(self.spin_min_rand, f"{name}_rand_min")
        set_qt_tips(self.spin_min_rand, "Minimum random file count.")

        self.spin_max_rand = QSpinBox(prefix="Max: ")
        set_qt_name(self.spin_max_rand, f"{name}_rand_max")
        set_qt_tips(self.spin_max_rand, "Maximum random file count.")

        self.spin_min_rand.valueChanged.connect(self.spin_max_rand.setMinimum)
        self.spin_max_rand.valueChanged.connect(self.spin_min_rand.setMaximum)

        self.radio_fixed.toggled.connect(self.spin_fixed.setEnabled)
        self.radio_fixed.toggled.connect(lambda c: self.spin_min_rand.setDisabled(c))
        self.radio_fixed.toggled.connect(lambda c: self.spin_max_rand.setDisabled(c))

        # Initial state
        self.spin_min_rand.setDisabled(True)
        self.spin_max_rand.setDisabled(True)

        layout = QGridLayout(self)
        layout.addWidget(self.radio_fixed, 0, 0)
        layout.addWidget(self.spin_fixed, 0, 1)
        layout.addWidget(self.radio_rand, 1, 0)
        layout.addWidget(self.spin_min_rand, 1, 1)
        layout.addWidget(self.spin_max_rand, 2, 1)

    def get_config(self) -> FilecountModel:
        """Return clean data for the config."""
        return FilecountModel(
            count=self.spin_fixed.value(),
            is_rand_enabled=self.radio_rand.isChecked(),
            rand_min=self.spin_min_rand.value(),
            rand_max=self.spin_max_rand.value(),
        )


class FolderCreatorWidget(BaseGroupBox):
    """Handles logic for creating folders."""

    def __init__(self) -> None:
        """Initialize the create folders widget."""
        name = "folder"
        super().__init__("Create Folders", name, checkable=True)

        self.spinbox_folder_count = QSpinBox(suffix=" folders")
        set_qt_name(self.spinbox_folder_count, f"{name}_count")
        set_qt_tips(self.spinbox_folder_count, "Number of folders to create.")

        self.lineedit_folder_name = QLineEdit(placeholderText="Ex: Random_Files", clearButtonEnabled=True)
        set_qt_name(self.lineedit_folder_name, f"{name}_name")
        set_qt_tips(self.lineedit_folder_name, "Template for naming created folders.")

        self.chk_unique_folders = QCheckBox("Ensure unique folders")
        set_qt_name(self.chk_unique_folders, f"{name}_unique")
        set_qt_tips(self.chk_unique_folders, "If checked, created folder names will have unique files.")

        layout = QVBoxLayout(self)
        layout.addWidget(self.spinbox_folder_count)
        layout.addWidget(self.lineedit_folder_name)
        layout.addWidget(self.chk_unique_folders)

    def get_config(self) -> DirectoryModel:
        """Return clean data for the config."""
        return DirectoryModel(
            is_enabled=self.isChecked(),
            is_unique=self.chk_unique_folders.isChecked(),
            name=self.lineedit_folder_name.text(),
            count=self.spinbox_folder_count.value() if self.isChecked() else 1,
        )


class FilenameWidget(BaseGroupBox):
    """Handles logic for filename template settings."""

    def __init__(self) -> None:
        """Initialize the filename template settings widget."""
        name: str = "filename"
        super().__init__("Filename", name)

        self.edit_template = QLineEdit("{original}", placeholderText="Ex: {original}_{index}", clearButtonEnabled=True)
        set_qt_name(self.edit_template, f"{name}_template")
        set_qt_tips(self.edit_template, "Template for renaming files. Use the 'Insert Tag' button to add tags.")

        self.menu_template = QMenu(title="Tags")
        set_qt_name(self.menu_template, f"{name}_template_menu")
        set_qt_tips(self.menu_template, "Select a tag to insert into the filename template.")

        self.btn_template = QPushButton("Insert tag")
        self.btn_template.setMenu(self.menu_template)
        set_qt_name(self.btn_template, f"{name}_template_button")
        set_qt_tips(self.btn_template, "Insert a tag into the template at the cursor position.")

        for lbl in FilenameTemplate:
            action = self.menu_template.addAction(lbl)
            action.triggered.connect(lambda _, tag=lbl: self.insert_tag(tag))

        layout = QHBoxLayout(self)
        layout.addWidget(self.edit_template)
        layout.addWidget(self.btn_template)

    @Slot(str)
    def insert_tag(self, tag: str) -> None:
        """Insert a tag into the template at the cursor position."""
        self.edit_template.insert(tag)
        self.edit_template.setFocus()

    def get_config(self) -> FilenameModel:
        """Return clean data for the config."""
        val = self.edit_template.text() or "{original}"
        return FilenameModel(template=val)


class TransferModeWidget(BaseGroupBox):
    """Handles logic for mode settings."""

    def __init__(self) -> None:
        """Initialize the mode settings widget."""
        name = "transfermode"
        super().__init__("Transfer Mode", name)

        self.combo_mode = QComboBox()
        self.combo_mode.addItems(get_available_transfer_modes())
        set_qt_name(self.combo_mode, f"{name}_mode")
        set_qt_tips(self.combo_mode, "Select the transfer mode to use.")

        layout = QVBoxLayout(self)
        layout.addWidget(self.combo_mode)

    def get_config(self) -> TransferModeModel:
        """Return clean data for the config."""
        return TransferModeModel(transfer_mode=self.combo_mode.currentText())


class ListIncludeExcludeFilterWidget(BaseGroupBox):
    """Handles the Include/Exclude pattern for Keywords and Extensions."""

    def __init__(self, title: str, name: str) -> None:
        """Initialize the dual list widget."""
        super().__init__(title, name, checkable=True)

        title_lower = self.title().casefold()

        self.filter_edit = QLineEdit(placeholderText="comma,separated,items", clearButtonEnabled=True)
        set_qt_name(self.filter_edit, f"{name}_text")
        set_qt_tips(self.filter_edit, f"Enter {title_lower} separated by commas.")

        self.filter_include_radio = QRadioButton("Include")
        self.filter_include_radio.setChecked(True)
        set_qt_name(self.filter_include_radio, f"{name}_include")
        set_qt_tips(self.filter_include_radio, f"Include only items matching the {title_lower} filter.")

        self.filter_exclude_radio = QRadioButton("Exclude")
        set_qt_name(self.filter_exclude_radio, f"{name}_exclude")
        set_qt_tips(self.filter_exclude_radio, f"Exclude items matching the {title_lower} filter.")

        layout = QHBoxLayout(self)
        layout.addWidget(self.filter_edit)
        layout.addWidget(self.filter_include_radio)
        layout.addWidget(self.filter_exclude_radio)

    def get_config(self) -> ListIncludeExcludeModel:
        """Return clean data for the config."""
        return ListIncludeExcludeModel(
            is_enabled=self.isChecked(),
            should_include=self.filter_include_radio.isChecked(),
            text=self.filter_edit.text(),
        )


class ExtensionsFilterWidget(ListIncludeExcludeFilterWidget):
    """Handles the Include/Exclude pattern for file extensions."""

    def __init__(self) -> None:
        """Initialize the extensions filter widget."""
        super().__init__("Extensions", "extension")


class KeywordsFilterWidget(ListIncludeExcludeFilterWidget):
    """Handles the Include/Exclude pattern for keywords."""

    def __init__(self) -> None:
        """Initialize the keywords filter widget."""
        super().__init__("Keywords", "keyword")


class MinMaxFilterWidget(BaseGroupBox):
    """Handles logic for ranges (Min/Max), e.g., Size or Duration."""

    def __init__(self, title: str, name: str, items: Sequence[str | ByteUnit | TimeUnit]) -> None:
        """Initialize the range filter widget."""
        super().__init__(title, name, checkable=True)

        self.spin_min = QDoubleSpinBox(prefix="Min ")
        set_qt_name(self.spin_min, f"{name}_minimum")
        set_qt_tips(self.spin_min, f"Minimum value for the {name} filter.")

        self.spin_max = QDoubleSpinBox(prefix="Max ")
        set_qt_name(self.spin_max, f"{name}_maximum")
        set_qt_tips(self.spin_max, f"Maximum value for the {name} filter.")

        self.spin_min.valueChanged.connect(self.spin_max.setMinimum)
        self.spin_max.valueChanged.connect(self.spin_min.setMaximum)

        self.combo_unit = QComboBox()
        self.combo_unit.addItems(items)
        set_qt_name(self.combo_unit, f"{name}_unit")
        set_qt_tips(self.combo_unit, f"Unit multiplier for the {name} filter.")

        layout = QHBoxLayout(self)
        layout.addWidget(self.spin_min)
        layout.addWidget(self.spin_max)
        layout.addWidget(self.combo_unit)

    def get_config(self) -> MinMaxModel:
        """Return clean data for the config."""
        return MinMaxModel(
            is_enabled=self.isChecked(),
            minimum=self.spin_min.value(),
            maximum=self.spin_max.value(),
            unit=self.combo_unit.currentText(),
        )


class SizeFilterWidget(MinMaxFilterWidget):
    """Handles logic for size range filter."""

    def __init__(self) -> None:
        """Initialize the size filter widget."""
        super().__init__("File Size", "filesize", tuple(ByteUnit))


class DurationFilterWidget(MinMaxFilterWidget):
    """Handles logic for duration range filter."""

    def __init__(self) -> None:
        """Initialize the duration filter widget."""
        super().__init__("Duration", "duration", tuple(TimeUnit))


class SizeLimitWidget(BaseGroupBox):
    """Handles logic for output folder size limit."""

    def __init__(self, title: str, name: str, items: Sequence[str | ByteUnit | TimeUnit]) -> None:
        """Initialize the size limit widget."""
        super().__init__(title, name, checkable=True)

        self.spin_size = QDoubleSpinBox(prefix="Size ")
        self.spin_size.setSpecialValueText("Unlimited")
        set_qt_name(self.spin_size, f"{name}_size")
        set_qt_tips(self.spin_size, f"{title} value for the size limit.")

        self.combo_unit = QComboBox()
        self.combo_unit.addItems(items)
        set_qt_name(self.combo_unit, f"{name}_unit")
        set_qt_tips(self.combo_unit, f"Unit multiplier for the {title} filter.")

        layout = QHBoxLayout(self)
        layout.addWidget(self.spin_size)
        layout.addWidget(self.combo_unit)

    def get_config(self) -> SizeLimitModel:
        """Return clean data for the config."""
        return SizeLimitModel(
            is_enabled=self.isChecked(),
            size_limit=self.spin_size.value(),
            unit=self.combo_unit.currentText(),
        )


class FolderSizeLimitWidget(SizeLimitWidget):
    """Handles logic for per-folder size limit."""

    def __init__(self) -> None:
        """Initialize the folder size limit widget."""
        super().__init__("Max Folder Size", "folder_size_limit", tuple(ByteUnit))


class TotalSizeLimitWidget(SizeLimitWidget):
    """Handles logic for total size limit across all folders."""

    def __init__(self) -> None:
        """Initialize the total size limit widget."""
        super().__init__("Max Total Size", "total_size_limit", tuple(ByteUnit))


class OptionsWidget(BaseGroupBox):
    """Handles logic for miscellaneous options."""

    def __init__(self) -> None:
        """Initialize the options widget."""
        name = "options"
        super().__init__("Options", name)

        self.spin_max_per_folder = QSpinBox()
        self.spin_max_per_folder.setSpecialValueText("Unlimited")
        set_qt_name(self.spin_max_per_folder, f"{name}_max_per_folder")
        set_qt_tips(self.spin_max_per_folder, "Maximum number of files allowed per input folder. 0 for unlimited.")

        self.chk_follow_symlink = QCheckBox()
        set_qt_name(self.chk_follow_symlink, f"{name}_should_follow_symlink")
        set_qt_tips(self.chk_follow_symlink, "If checked, symbolic links will be followed during file traversal.")

        self.chk_dry_run = QCheckBox()
        set_qt_name(self.chk_dry_run, f"{name}_dry_run")
        set_qt_tips(self.chk_dry_run, "If checked, no files will actually be copied.")

        layout = QFormLayout(self)
        layout.addRow("Max from one folder", self.spin_max_per_folder)
        layout.addRow("Follow symbolic links", self.chk_follow_symlink)
        layout.addRow("Dry run (simulation)", self.chk_dry_run)

    def get_config(self) -> OptionsModel:
        """Return clean data for the config."""
        return OptionsModel(
            max_per_folder=self.spin_max_per_folder.value(),
            should_follow_symlink=self.chk_follow_symlink.isChecked(),
            is_dry_run=self.chk_dry_run.isChecked(),
        )


class ProgressWidget(QWidget):
    """Progress bars and execution controls."""

    def __init__(self) -> None:
        """Post-initialize the progress widget."""
        super().__init__()
        name = "progress"
        set_qt_name(self, name)

        self.progbar_total = QProgressBar(textVisible=True)
        set_qt_name(self.progbar_total, f"{name}_total")
        set_qt_tips(self.progbar_total, "Total progress bar, max is set at number of output folders.")

        self.progbar_dir = QProgressBar(textVisible=True)
        set_qt_name(self.progbar_dir, f"{name}_dir")
        set_qt_tips(self.progbar_dir, "Current folder progress bar, max is set at number of files to copy.")

        layout = QFormLayout(self)
        layout.addRow("Total", self.progbar_total)
        layout.addRow("Folder", self.progbar_dir)

    @Slot()
    def update_total_prog(self) -> None:
        """Update the total progress bar."""
        self.progbar_total.setValue(self.progbar_total.value() + 1)

    def reset(self) -> None:
        """Reset progress bars."""
        self.progbar_total.setValue(0)
        self.progbar_dir.setValue(0)


class LoggingWidget(QWidget):
    """Logging widget."""

    def __init__(self) -> None:
        """Post-initialize the logging widget."""
        super().__init__()
        name = "logging"
        set_qt_name(self, name)

        self.textbrowser_log = QTextBrowser()
        self.textbrowser_log.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
        set_qt_name(self.textbrowser_log, f"{name}_log")
        set_qt_tips(self.textbrowser_log, "Log for output messages.")

        layout = QHBoxLayout(self)
        layout.addWidget(self.textbrowser_log)


class ProgressBinder(QObject):
    """Class for binding progress widgets."""

    count: ClassVar[Signal] = Signal(int)
    finished: ClassVar[Signal] = Signal()

    def __init__(self, progressw: ProgressWidget, loggingw: LoggingWidget) -> None:
        """Initialize the ProgressBinder."""
        super().__init__()
        self.progress = progressw
        self.logging = loggingw

    def bind(self, signals: WorkerSignals) -> None:
        """Bind worker signals to progress widget."""
        signals.progress_total.connect(self.progress.progbar_total.setMaximum)
        signals.count_total.connect(self.progress.update_total_prog)
        signals.progress.connect(self.progress.progbar_dir.setMaximum)
        signals.log.connect(self.logging.textbrowser_log.append)
        signals.count.connect(self.on_count)
        signals.finished.connect(self.finished.emit)

    @Slot(int)
    def on_count(self, count: int) -> None:
        """Handle directory progress count update."""
        self.progress.progbar_dir.setValue(count)
        self.count.emit(count)



"""Main module."""

from dataclasses import dataclass, field

from PySide6.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget

from ..config import ConfigModel
from .components import (
    DestPathSelectorWidget,
    DurationFilterWidget,
    ExtensionsFilterWidget,
    FileCountWidget,
    FilenameWidget,
    FolderCreatorWidget,
    FolderSizeLimitWidget,
    KeywordsFilterWidget,
    ListIncludeExcludeFilterWidget,
    LoggingWidget,
    MinMaxFilterWidget,
    OptionsWidget,
    PathSelectorWidget,
    ProgressWidget,
    RootPathSelectorWidget,
    SizeFilterWidget,
    SizeLimitWidget,
    TotalSizeLimitWidget,
    TransferModeWidget,
)


@dataclass(slots=True)
class UIBuilder:
    """Main Widget Builder."""

    root: PathSelectorWidget = field(default_factory=RootPathSelectorWidget)
    dest: PathSelectorWidget = field(default_factory=DestPathSelectorWidget)
    filecount: FileCountWidget = field(default_factory=FileCountWidget)
    folders: FolderCreatorWidget = field(default_factory=FolderCreatorWidget)
    filename: FilenameWidget = field(default_factory=FilenameWidget)
    transfermode: TransferModeWidget = field(default_factory=TransferModeWidget)
    keywords: ListIncludeExcludeFilterWidget = field(default_factory=KeywordsFilterWidget)
    extensions: ListIncludeExcludeFilterWidget = field(default_factory=ExtensionsFilterWidget)
    filesize: MinMaxFilterWidget = field(default_factory=SizeFilterWidget)
    duration: MinMaxFilterWidget = field(default_factory=DurationFilterWidget)
    folder_size_limit: SizeLimitWidget = field(default_factory=FolderSizeLimitWidget)
    total_size_limit: SizeLimitWidget = field(default_factory=TotalSizeLimitWidget)
    options: OptionsWidget = field(default_factory=OptionsWidget)
    progress: ProgressWidget = field(default_factory=ProgressWidget)
    logging: LoggingWidget = field(default_factory=LoggingWidget)

    def build_layout(self) -> QVBoxLayout:
        """Set up the main UI layouts."""
        path_widget = QWidget()
        path_layout = QVBoxLayout(path_widget)
        path_layout.addWidget(self.root)
        path_layout.addWidget(self.dest)

        # Left widget
        output_widget = QWidget()
        output_layout = QHBoxLayout(output_widget)
        output_layout.addWidget(self.filecount)
        output_layout.addWidget(self.transfermode)
        output_layout.addWidget(self.folders)
        output_layout.addWidget(self.filename)

        filter_widget = QWidget()
        filter_layout = QVBoxLayout(filter_widget)
        filter_layout.addWidget(self.keywords)
        filter_layout.addWidget(self.extensions)

        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        left_layout.addWidget(output_widget)
        left_layout.addWidget(filter_widget)

        # Right widget
        size_widget = QWidget()
        size_layout = QHBoxLayout(size_widget)
        size_layout.addWidget(self.filesize)
        size_layout.addWidget(self.duration)

        size_limit_widget = QWidget()
        size_limit_layout = QHBoxLayout(size_limit_widget)
        size_limit_layout.addWidget(self.folder_size_limit)
        size_limit_layout.addWidget(self.total_size_limit)

        options_widget = QWidget()
        options_layout = QHBoxLayout(options_widget)
        options_layout.addWidget(self.options)

        right_widget = QWidget()
        right_layout = QVBoxLayout(right_widget)
        right_layout.addWidget(size_widget)
        right_layout.addWidget(size_limit_widget)
        right_layout.addWidget(options_widget)

        # Body widget
        body_widget = QWidget()
        body_layout = QHBoxLayout(body_widget)
        body_layout.addWidget(left_widget)
        body_layout.addWidget(right_widget)

        # Assemble Main Layout
        main_layout = QVBoxLayout()
        main_layout.addWidget(path_widget)
        main_layout.addWidget(body_widget)
        main_layout.addWidget(self.logging)
        main_layout.addWidget(self.progress)
        return main_layout

    def get_config(self) -> ConfigModel:
        """Get the current configuration from all widgets."""
        return ConfigModel(
            root=self.root.get_config(),
            dest=self.dest.get_config(),
            filecount=self.filecount.get_config(),
            folder=self.folders.get_config(),
            filename=self.filename.get_config(),
            transfermode=self.transfermode.get_config(),
            keyword=self.keywords.get_config(),
            extension=self.extensions.get_config(),
            filesize=self.filesize.get_config(),
            duration=self.duration.get_config(),
            folder_size_limit=self.folder_size_limit.get_config(),
            total_size_limit=self.total_size_limit.get_config(),
            options=self.options.get_config(),
        )



"""General paths class and utilities."""

import os
from dataclasses import dataclass
from typing import ClassVar

import fspachinko


@dataclass(slots=True)
class Paths:
    """Dataclass for general directories used."""

    pkg: ClassVar[str] = os.path.dirname(fspachinko.__file__)
    data: ClassVar[str] = os.path.join(pkg, "_data")
    icons: ClassVar[str] = os.path.join(data, "icons")
    configs: ClassVar[str] = os.path.join(data, "configs")
    profiles: ClassVar[str] = os.path.join(data, "gui_profiles")

    @classmethod
    def icon(cls, path: str) -> str:
        """Get the full path to an icon."""
        return os.path.join(cls.icons, path)

    @classmethod
    def config(cls, path: str) -> str:
        """Get the full path to a config file."""
        return os.path.join(cls.configs, path)

    @classmethod
    def profile(cls, path: str) -> str:
        """Get the full path to a profile file."""
        return os.path.join(cls.profiles, path)



"""Provider for current date and time."""

from dataclasses import dataclass, field
from datetime import UTC, datetime

from .constants import DateTimeFormat


@dataclass(slots=True)
class DateTimeStamp:
    """Provider for current date and time."""

    _now: datetime = field(init=False)
    date: str = ""
    time: str = ""
    date_time: str = ""
    date_time_report_str: str = ""

    def __post_init__(self) -> None:
        """Post-initialization tasks."""
        self.refresh()

    def refresh(self) -> None:
        """Refresh the current date and time."""
        self._now = datetime.now(tz=UTC)
        self.date = self._now.strftime(DateTimeFormat.DATE)
        self.time = self._now.strftime(DateTimeFormat.TIME)
        self.date_time = f"{self.date}--{self.time}"
        self.date_time_report_str = self._now.strftime(DateTimeFormat.DATETIME)



"""Main entry point for CLI."""

import logging

from ..utils import initialize_logging
from .app import app

logger = logging.getLogger(__name__)


def main() -> None:
    """Enter CLI."""
    initialize_logging()
    logger.info("Start: fspachinko CLI")
    app()


if __name__ == "__main__":
    main()



"""Random file system navigator."""

import logging
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from random import Random

    from .quota import DiversityQuota
    from .validator import FileValidator

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class FSWalker(ABC):
    """Abstract file system walker."""

    @abstractmethod
    def reset(self) -> None:
        """Reset the walker for a new batch."""

    @abstractmethod
    def walk(self) -> Iterator[FSEntry]:
        """Generate candidates for a given directory."""


@dataclass(slots=True)
class FSEntry:
    """Lightweight wrapper for os.DirEntry with only path and name."""

    path: str
    stem: str
    ext: str
    size: int

    @classmethod
    def from_direntry(cls, e: os.DirEntry) -> FSEntry:
        """Create a lightweight FSEntry from an os.DirEntry."""
        stem, ext = os.path.splitext(e.name)
        return cls(
            path=e.path,
            stem=stem,
            ext=ext,
            size=e.stat().st_size,
        )

    def __hash__(self) -> int:
        """Return the hash based on the file path."""
        return hash(self.path)

    def __fspath__(self) -> str:
        """Return the file system path representation."""
        return self.path


@dataclass(slots=True)
class FSPachinkoPin:
    """Represents a 'pin' on the Pachinko board."""

    path: str
    subdirs: list[str] = field(default_factory=list)
    files: list[FSEntry] = field(default_factory=list)
    is_scanned: bool = False
    is_exhausted: bool = False


@dataclass(slots=True)
class PachinkoFSWalker(FSWalker):
    """Simulates a Pachinko machine.

    For every file needed, we 'drop' a search cursor from the Root.
    It bounces randomly down directory paths until it settles on a file.
    """

    root: str
    quota: DiversityQuota
    validator: FileValidator
    rng: Random
    should_follow_symlink: bool
    board: dict[str, FSPachinkoPin] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Initialize the board with the root pin."""
        self.reset()

    def reset(self) -> None:
        """Reset the walker and quota for a new batch."""
        self.board.clear()
        self.board[self.root] = FSPachinkoPin(path=self.root)

    def walk(self) -> Iterator[FSEntry]:
        """Continuously drop balls until the board is empty."""
        while not self.board[self.root].is_exhausted:
            if (entry := self.drop()) is not None:
                yield entry

    def drop(self) -> FSEntry | None:
        """Drop a ball from the root."""
        current_path = self.root

        while True:
            pin = self.board[current_path]

            if pin.is_exhausted:
                return None

            if not pin.is_scanned:
                self.scan(pin)

            valid_subdirs = self.get_valid_subdirs(pin)
            valid_files = pin.files

            has_subdirs, has_files = bool(valid_subdirs), bool(valid_files)

            if not (has_subdirs or has_files):
                self.mark_exhausted(pin)
                return None

            if self.should_descend(has_subdirs=has_subdirs, has_files=has_files):
                current_path = self.rng.choice(valid_subdirs)
                continue

            return pin.files.pop()

    def get_valid_subdirs(self, pin: FSPachinkoPin) -> list[str]:
        """Get valid subdirectories for a given pin."""
        valid = [d for d in pin.subdirs if not self.quota.is_dir_locked(d) and not self.board[d].is_exhausted]
        pin.subdirs = valid
        return valid

    def mark_exhausted(self, pin: FSPachinkoPin) -> None:
        """Mark a pin and all its subdirs as exhausted."""
        currpath = pin.path
        pin.is_exhausted = True
        self.quota.lock_dir(currpath)

    def should_descend(self, *, has_subdirs: bool, has_files: bool) -> bool:
        """Decide whether to descend into a subdir or select a file."""
        if has_subdirs and has_files:
            return self.rng.choice([True, False])
        return has_subdirs

    def scan(self, pin: FSPachinkoPin) -> None:
        """Only look at the OS file system when a ball hits a specific folder for the first time."""
        is_valid = self.validator.is_valid
        subdirs = []
        files = []
        followlinks = self.should_follow_symlink

        try:
            with os.scandir(pin.path) as it:
                for e in it:
                    try:
                        if e.is_symlink() and not followlinks:
                            continue

                        if e.is_dir(follow_symlinks=followlinks):
                            dirpath = e.path
                            subdirs.append(dirpath)
                            if dirpath not in self.board:
                                self.board[dirpath] = FSPachinkoPin(path=dirpath)
                        elif e.is_file(follow_symlinks=followlinks):
                            fsentry = FSEntry.from_direntry(e)
                            if is_valid(fsentry):
                                files.append(fsentry)
                    except OSError:
                        continue
        except OSError:
            pin.is_exhausted = True
            return

        if subdirs:
            self.rng.shuffle(subdirs)
        if files:
            self.rng.shuffle(files)

        pin.subdirs = subdirs
        pin.files = files
        pin.is_scanned = True
